import os
import signal
import subprocess
import threading
import time
import atexit
import sys


class PythonExecutor:
    def __init__(self, timeout=6):
        self.timeout = timeout
        self.stop_event = threading.Event()
        self.processes = []  # track all processes

        # Register cleanup at exit
        atexit.register(self.cleanup)

        # Handle Ctrl+C / kill signals
        signal.signal(signal.SIGINT, self.handle_exit)
        signal.signal(signal.SIGTERM, self.handle_exit)

    def read_stream(self, stream, stream_name):
        try:
            for line in stream:
                if self.stop_event.is_set():
                    break
                if stream_name == "stdout":
                    self.output[0].append(line)
                else:
                    self.output[1].append(line)
                # self.output.append((stream_name, line))
        finally:
            stream.close()

    def timeout_handler(self, process):
        time.sleep(self.timeout)
        if not self.stop_event.is_set():
            self.stop_event.set()
            self.kill_process(process)

    def kill_process(self, process):
        if process and process.poll() is None:
            try:
                # Kill the whole process group
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                os.killpg(os.getpgid(process.pid), signal.SIGKILL)
            except ProcessLookupError:
                pass  # already gone

    def executeSelfExitingCommand(self,command,env=None,cwd=None):

        process = subprocess.Popen(command,env=env,cwd=cwd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        
        stdout, stderr = process.communicate()

        if stdout:
            lines = stdout.decode()
            lines = lines.split("\n")
            for line in lines:
                if len(line.strip()) > 0:
                    self.output[0].append(line)

        if stderr:
            lines = stderr.decode()
            lines = lines.split("\n")
            for line in lines:
                if len(line.strip()) > 0:
                    self.output[1].append(line)

    def runCommand(self, command, cwd,env=None, self_exit=True):
        self.output = [[],[]]
        self.stop_event = threading.Event()

        # print("self exist ", "#"*5, self_exit)

        if self_exit:
            self.executeSelfExitingCommand(command,env,cwd)
        else:
            process = subprocess.Popen(
                command,
                cwd=cwd,
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
                preexec_fn=os.setsid  # create new process group
            )
            self.processes.append(process)

            stdout_thread = threading.Thread(
                target=self.read_stream,
                args=(process.stdout, 'stdout'),
                daemon=True
            )
            stderr_thread = threading.Thread(
                target=self.read_stream,
                args=(process.stderr, 'stderr'),
                daemon=True
            )
            timeout_thread = threading.Thread(
                target=self.timeout_handler,
                args=(process,),
                daemon=True
            )

            stdout_thread.start()
            stderr_thread.start()
            timeout_thread.start()

            process.wait()
            self.stop_event.set()

            stdout_thread.join(timeout=2)
            stderr_thread.join(timeout=2)
            timeout_thread.join(timeout=1)

            self.kill_process(process)

        return self.output

    def cleanup(self):
        """Kill all subprocesses at exit"""
        for process in self.processes:
            self.kill_process(process)

    def handle_exit(self, signum, frame):
        """Handle Ctrl+C or kill signals"""
        print(f"\nReceived signal {signum}, cleaning up...")
        self.cleanup()
        sys.exit(1)



"""move template code to target file"""
import os
import json
from pathlib import Path

from ....utils.state import clearProjectState, copyFileToState, getOrCreateProjectSBId, updateState

from ....parsers.javascript_typescript.js_var_names import get_variable_name_and_type

from .merge_js_entry import mergeEntryFiles
from ....cli_output.output import StatusManager
from ....parsers.javascript_typescript.jsParser import JsTxParser
from ....src.Express.extract.reverseImport import convertToImportStatements


parser = JsTxParser()
home = str(Path.home())
test_environment = f"{home}/.sb/environment/express"

def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_paths.append(absolute_path)
    return file_paths


def processPackageDotJson(path):
    package_path = f"{path}/package.json"
    if not os.path.exists(package_path):
        raise ValueError("Cannot find package.json in test environment")
    
    with open(package_path,"r") as file:
        data = json.loads(file.read())
        return [
            data['main'],
            data['dependencies'],
            data['dev-dependencies'] if "dev-dependencies" in data.keys() else None
        ]
    
    raise ValueError("Invalid package.json content")

def writeExpressCodeToFile(dest,import_deps,chunks):
    
    with open(dest,"w") as file:
        code = ""
        if len(import_deps) > 0:
            import_statements = convertToImportStatements(import_deps)
            code += "\n".join(import_statements)

        if len(chunks) > 0:
            code += "\n\n"
            code += "\n\n".join(chunks)

        file.write(code)

async def moveTemplateFilesToTargetProject(destination_addr):
    logger = StatusManager()

    logger.start_status("Moving Template Files to target project")

    template_files = getTemplateFileNames(test_environment)
    entry_file, dependencies, devDependencies = processPackageDotJson(test_environment)

    destination_deploy_addr = f"{destination_addr}/speedbuild" #TODO : make dynamuc

    # managing state
    project_id = getOrCreateProjectSBId(destination_addr)
    clearProjectState(project_id) # clear project state 

    
    
    template_files_in_destination_folder = []
    file_written = []

    for file in template_files:
        if os.path.basename(file) not in ['package.json',"package-lock.json",entry_file]:
            logger.update_status(f"Moving : {os.path.basename(file)}")
            file_path = file.replace(test_environment,"").lstrip("/")
            # full_dest_path = f"{destination_addr}/{file_path}"
            full_dest_path = f"{destination_deploy_addr}/{file_path}"

            file_state_key = full_dest_path[len(destination_addr):].lstrip("/")

            action = "update" # state file action

            if not os.path.exists(full_dest_path):
                os.makedirs(os.path.dirname(full_dest_path),exist_ok=True)
                chunks = []
                import_deps = {}
                chunk_var_mapping = {}
                action = "create" # update state file action to create if file does not exist
            else:
                _,chunks,chunk_var_mapping,import_deps = await parser.parse_code(full_dest_path)
            

            # print("file state file is ",file_state_key)

            # saving current destination file state
            if project_id:
                if full_dest_path not in file_written:
                    # copy file to prev state
                    copyFileToState(project_id,full_dest_path,destination_addr,"prev")
                    
                    # update state.json to state file updated
                    updateState(project_id,action,file_state_key)

                    # add to file_written
                    file_written.append(full_dest_path)


            # get template contents
            _,template_chunks,_,template_import_deps = await parser.parse_code(f"{test_environment}/{file_path}")
            
            # merge chunks
            for chunk in template_chunks:

                if chunk not in chunks:
                    chunk_name = get_variable_name_and_type(chunk) # returns a list : [type, var_name]
      
                    if chunk_name != None and chunk_name[1] in chunk_var_mapping and chunk != chunk_var_mapping[chunk_name[1]]:
                        index = chunks.index(chunk_var_mapping[chunk_name[1]])
                        conflictCode = f"<<<<<<< SpeedBuild update \n{chunk_var_mapping[chunk_name[1]]}\n=======\n{chunk}\n>>>>>>>"
                        chunks[index] = conflictCode
                    else:
                        chunks.append(chunk)

            # merge import 
            # TODO : import issue here
            for importStatement in template_import_deps:
                if importStatement not in import_deps:
                    import_deps[importStatement]= template_import_deps[importStatement]

            # write code to destination
            template_files_in_destination_folder.append(full_dest_path)
            writeExpressCodeToFile(full_dest_path,import_deps,chunks)

            # state after write
            if project_id:
                # copy file to last_sb_update state
                copyFileToState(project_id,full_dest_path,destination_addr,"last_sb_update")

    # merge project entry files
    logger.update_status("Processing package.json")
    dest_entry_file, _, _ = processPackageDotJson(destination_addr)

    logger.stop_status()

    entry_file_path = f"{destination_addr}/{dest_entry_file}"
    action = "update" # state file action
    if project_id:
        # copy entry file to prev state
        copyFileToState(project_id,entry_file_path,destination_addr,"prev")

        file_state_key = entry_file_path[len(destination_addr):].lstrip("/")
        
        # update state.json to state file updated
        updateState(project_id,action,file_state_key)

    import_deps,chunks = await mergeEntryFiles(
        entry_file_path,
        f"{test_environment}/{entry_file}",
        template_files_in_destination_folder,
        logger
    )

    # write entry 
    writeExpressCodeToFile(entry_file_path,import_deps,chunks)

    # state after write
    if project_id:
        # copy file to last_sb_update state
        copyFileToState(project_id,entry_file_path,destination_addr,"last_sb_update")

    # process package.json
    dest_package_json = f"{destination_addr}/package.json"

    if project_id:
        # copy entry file to prev state
        copyFileToState(project_id,dest_package_json,destination_addr,"prev")

        file_state_key = dest_package_json[len(destination_addr):].lstrip("/")
        
        # update state.json to state file updated
        updateState(project_id,action,file_state_key)

    if not os.path.exists(dest_package_json):
        raise ValueError("can't find target project package.json")
    
    with open(dest_package_json,"r+") as file:
        data = json.loads(file.read())

        if dependencies is not None and len(dependencies.keys()) > 0:
            for dep in dependencies:
                if dep not in data['dependencies']:
                    data['dependencies'][dep] = dependencies[dep]

        if devDependencies is not None and len(devDependencies.keys()) > 0:
            for dep in devDependencies:
                if dep not in data['devDependencies']:
                    data['devDependencies'][dep] = devDependencies[dep]

        # clear file and write new file content
        file.seek(0)
        file.truncate()
        file.write(json.dumps(data,indent=4))
    
    if project_id:
        # copy file to last_sb_update state
        copyFileToState(project_id,dest_package_json,destination_addr,"last_sb_update")
    
    logger.print_message("Template files moved to target project")


# TODO : remove state repetition from this file


import os
import ast

from .src.Django.utils.manage_url import merge_urls
from .parsers.python.parser import PythonBlockParser
from .src.Django.utils.django_utils import django_defaults
from .src.Django.utils.var_utils import get_assigned_variables

parser = PythonBlockParser()

def getProjectFiles(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root or "__pycache__" in root or "migrations" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_name = absolute_path.replace(path,"").lstrip("/")
            if file_name not in ["__init__.py","apps.py"]:
                file_paths.append(file_name)
    return file_paths

def getImportsFromChunks(chunks):
    imports = []
    code = []

    for chunk in chunks:
        if chunk.startswith("import ") or chunk.startswith("from "):
            imports.append(chunk)
        else:
            code.append(chunk)

    return [imports,code]


def getFileData(file_path,context=None):

    if not os.path.exists(file_path):
        return [[],[],{}]
    
    with open(file_path,"r") as file:
        data = file.read()
        if context is not None:
            # get source project context
            source_project_name,source_app_name,project_name,app_name = getProjectContext(context)

            # match and replace
            data = data.replace(source_project_name,project_name).replace(source_app_name,app_name)

        chunks = parser.parse_code(data)

        imports, chunks = getImportsFromChunks(chunks)
        var_mappings = {}

        for chunk in chunks:
            var_name = get_assigned_variables(chunk,True)
            if isinstance(var_name,str):
                var_mappings[var_name] = chunk

        return imports,chunks,var_mappings
    

def getProjectContext(context):
    return [
        context['source_project_name'],
        context['source_app_name'],
        context['project_name'],
        context['app_name']
    ]


def mergeFile(source,target_file,project_context):
    source_import,source_code,_ = getFileData(source,project_context)
    target_file_imports,target_file_code,target_file_var_mapping = getFileData(target_file)

    # merge imports
    for statement in source_import:
        if statement not in target_file_imports:
            target_file_imports.append(statement)

    # merge code
    for chunk in source_code:
        if chunk not in target_file_code:
            chunk_name = get_assigned_variables(chunk,True)

            if isinstance(chunk_name,str) and chunk_name in target_file_var_mapping.keys():
                target_file_chunk_with_chunk_name = target_file_var_mapping[chunk_name]

                if target_file_chunk_with_chunk_name != chunk:
                    target_file_chunk_pos = target_file_code.index(target_file_chunk_with_chunk_name)

                    conflictCode = f"<<<<<<< SpeedBuild update \n{target_file_chunk_with_chunk_name}\n=======\n{chunk}\n>>>>>>>"
                    target_file_code[target_file_chunk_pos] = conflictCode
            elif chunk not in target_file_code: 
                target_file_code.append(chunk)
    
    code = ""
    if len(target_file_imports) > 0:
        code = "\n".join(target_file_imports) + "\n\n"

    code += "\n\n".join(target_file_code)

    # write to file
    output_dir = os.path.dirname(target_file)
    if not os.path.exists(output_dir):
        print("making ", output_dir)
        os.makedirs(output_dir)

    with open(target_file,"w") as file:
        file.write(code)

def handleUrlFileMerge(source_url,target_url, project_context):

    # get source project context
    source_project_name,source_app_name,project_name,app_name = getProjectContext(project_context)

    with open(source_url,"r") as file:
        source_url_data = file.read()
        # replace project and app name in source url
        source_url_data = source_url_data.replace(source_project_name,project_name).replace(source_app_name,app_name).split("\n")

    if os.path.exists(target_url):
        with open(target_url,"r") as file:
            target_url_data = file.read().split("\n")

        code = merge_urls(source_url_data,target_url_data)
        # print(code)
    else:
        print("no url file")
        code = source_url_data
    
    with open(target_url,"w") as file:
        file.write(code)

def getProjectApps(apps):
    cleaned = []
    for app in apps:
        if not app.startswith("django.") and app != "speedbuild_app":
            cleaned.append(app)
    return cleaned

def mergeMiddleware(middleware, target_middleware):
    for i in range(len(middleware)):
        item = middleware[i]
        if item not in target_middleware:
            if i == 0:
                target_middleware.insert(0, item)
            elif i == len(middleware) - 1:
                target_middleware.append(item)
            else:
                added = False
                # Try to insert after the closest previous middleware already in target_middleware
                for x in reversed(middleware[:i]):
                    if x in target_middleware:
                        pos = target_middleware.index(x)
                        target_middleware.insert(pos + 1, item)
                        added = True
                        break
                if not added:
                    target_middleware.append(item)
    return target_middleware

def handleDjangoSettingsFile(settings_path,target_settings_file,project_context):

    # imports installed_apps, middleware configurations
    imports, _, code_var_mappings = getFileData(settings_path,project_context)
    target_imports, target_chunks, target_code_var_mappings = getFileData(target_settings_file)

    MIDDLEWARE = ast.literal_eval(code_var_mappings["MIDDLEWARE"].split("=", 1)[1].strip())
    INSTALLED_APPS = getProjectApps(ast.literal_eval(code_var_mappings["INSTALLED_APPS"].split("=", 1)[1].strip()))

    target_MIDDLEWARE = ast.literal_eval(target_code_var_mappings["MIDDLEWARE"].split("=", 1)[1].strip())
    target_INSTALLED_APPS = ast.literal_eval(target_code_var_mappings["INSTALLED_APPS"].split("=", 1)[1].strip())

    # TODO : add project app if its not already in installed apps
    installed_app_pos = target_chunks.index(target_code_var_mappings["INSTALLED_APPS"])
    middleware_pos = target_chunks.index(target_code_var_mappings["MIDDLEWARE"])

    for app in INSTALLED_APPS:
        if app not in target_INSTALLED_APPS:
            target_INSTALLED_APPS.append(app)

    MIDDLEWARE_merged = mergeMiddleware(MIDDLEWARE,target_MIDDLEWARE)

    target_INSTALLED_APPS = ",\n\t".join([f'"{app}"' for app in target_INSTALLED_APPS]) #match and replace
    MIDDLEWARE_merged = ",\n\t".join([f'"{middleware}"' for middleware in MIDDLEWARE_merged]) #match and replace

    target_chunks[installed_app_pos] = f"INSTALLED_APPS = [\n\t{target_INSTALLED_APPS}\n]"
    target_chunks[middleware_pos] = f"MIDDLEWARE = [\n\t{MIDDLEWARE_merged}\n]"

    for statement in imports: #match and replace
        if statement not in target_imports:
            target_imports.append(statement)

    for i in code_var_mappings.keys(): #match and replace
        if i not in django_defaults:
            chunk = code_var_mappings[i]
            if i in target_code_var_mappings.keys() and chunk != target_code_var_mappings[i]:
                index = target_chunks.index(target_code_var_mappings[i])
                conflictCode = f"<<<<<<< SpeedBuild update \n{target_code_var_mappings[i]}\n=======\n{chunk}\n>>>>>>>"
                target_chunks[index] = conflictCode
            else:
                target_chunks.append(chunk)

    merged_code = ""
    if len(target_imports) > 0:
        merged_code = "\n".join(target_imports) + "\n\n"

    merged_code += "\n\n".join(target_chunks)

    with open(target_settings_file,"w") as file:
        file.write(merged_code)

def moveFolderFiles(source_path,target_root,project_context):
    app_files = getProjectFiles(source_path)

    for file in app_files:
        if file == "urls.py":
            handleUrlFileMerge(f"{source_path}/urls.py",f"{target_root}/urls.py",project_context)
        elif file == "settings.py":
            handleDjangoSettingsFile(f"{source_path}/settings.py",f"{target_root}/settings.py",project_context)
        else:
            mergeFile(f"{source_path}/{file}",f"{target_root}/{file}",project_context)

def moveFilesToTargetProject(source_path,target_root,project_name,app_name):
    test_environment_app_name = "speedbuild_app"
    test_environment_project_name = "speedbuild_project"

    project_context = {
        "source_project_name":test_environment_project_name,
        "source_app_name":test_environment_app_name,
        "project_name":project_name,
        "app_name":app_name
    }

    # move app files
    moveFolderFiles(f"{source_path}/{test_environment_app_name}",f"{target_root}/{app_name}",project_context)

    # move project files
    moveFolderFiles(f"{source_path}/{test_environment_project_name}",f"{target_root}/{project_name}",project_context)

# moveFilesToTargetProject(
#     "/home/attah/.sb/environment/django/speedbuild_project",
#     "/home/attah/Documents/speedbuild/test_environment/mybookshop",
#     "mybookshop", #make this dynamic; find the folder with the settings.py file
#     "shop"
# )


from ..src.Django.utils.feature_dependencies import getBlockDependencies
from ..src.Django.utils.var_utils import get_assigned_variables


def removeDecorator(code):
    blocks = code.split("\n")
    index = None

    for i in range(len(blocks)):
        if blocks[i].startswith("@"):
            index = i
        else:
            break
    
    if index != None:
        code_without_decorator = blocks[index:]
        return "\n".join(code_without_decorator).strip()
    
    return "\n".join(blocks).strip()


def addOrUpdateCode(current_code,new_code,old_code=None):

    if "import" in new_code:
        current_code.insert(0,new_code)
        return current_code
    
    var_name = []
    for block in current_code:
        code = removeDecorator(block)
        if not (code.startswith("import ") or code.startswith("from ")):
            var_name.append(get_assigned_variables(block))

    if old_code == None:        
        dependencies = getBlockDependencies(new_code,current_code)
        try:
            max_index = max([var_name.index(i['imports']) for i in dependencies if "imports " not in i and i['imports'] in var_name])
            
            if max_index + 1 > len(var_name):
                current_code.append(new_code)
            else:
                current_code.insert(max_index+1, new_code)

        except:
            current_code.append(new_code)
            
        
    else:
        # get old_code name
        name = get_assigned_variables(removeDecorator(old_code))
        index = var_name.index(name)
        current_code[index] = new_code
    
    return current_code





from openai import OpenAI
from pydantic import BaseModel

from typing import List 

from .sb_agent import getLLMClient
from .task_prompt import prompt as system_prompt


memory = [
    {"role": "system", "content": system_prompt},
]

class SpeedBuildFormat(BaseModel):
    actions: List[str]

def makeLLMCall(user_input):
    
    model, client = getLLMClient()

    memory.append({"role": "user","content": user_input})
    completion = client.beta.chat.completions.parse(
        model=model,
        messages=memory,
        response_format=SpeedBuildFormat,
    )

     # Extract the assistant's response
    assistant_response = completion.choices[0].message.parsed #completion.choices[0].message.content

    # Add the assistant's response to memory
    memory.append({"role": "assistant", "content": f"{assistant_response.actions}"})

    return assistant_response

def query_splitter(query):
    response = makeLLMCall(query)
    print(response, " response is video")
    return response.actions


import os
import shutil
from pathlib import Path
from rich.prompt import Prompt

from ....src.Express.deploy.debug import debugTemplate

from .extractJs import handleExtraction
from .move_file import list_all_files, moveFile
from ....cli_output.output import StatusManager
from ....langgraph_agent.agents.indexer.indexJs import Indexer
from .handle_js_route import getRoutePathAndMethodsOrReturnNone
from ....parsers.javascript_typescript.jsParser import JsTxParser
from ....utils.template_update import checkTemplateForVersionUpdate


def clear_folder(folder_path):
    """
    Clear all contents of a specified folder, removing files, symlinks, and subdirectories.

    Args:
        folder_path (str): Path to the folder to be cleared.

    Returns:
        None

    Raises:
        OSError: If there are permission issues or other OS-level errors when deleting files.
        FileNotFoundError: If the specified folder path does not exist.

    Examples:
        >>> clear_folder('/path/to/folder')
        # All contents of the folder will be deleted
    """
    for file_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, file_name)
        if os.path.isfile(file_path) or os.path.islink(file_path):  
            os.remove(file_path)  # Delete files & symlinks
        elif os.path.isdir(file_path):  
            shutil.rmtree(file_path)  # Delete directories


async def startExpressExtraction(path,file_name,project_root):
    """
    Extracts and processes Express.js route handlers from JavaScript files.
    This function parses JavaScript code, extracts route handlers matching a specified path,
    processes dependencies, and organizes extracted files into appropriate folders.
    Args:
        path (str): The route path to extract (e.g., "/api/users")
        file_name (str): Path to the JavaScript source file
        project_root (str): Root directory of the project
    Returns:
        None
    Raises:
        ValueError: If the specified route path is not found in the source file
    The function performs the following steps:
    1. Parses JavaScript code to extract chunks and dependencies
    2. Processes and maps dependencies
    3. Locates and extracts the route handler matching the specified path
    4. Indexes and categorizes extracted files
    5. Moves files to appropriate folders based on categorization
    6. Creates a zip archive of the extracted template
    """
    parser = JsTxParser()
    logger = StatusManager()

    logger.print_message("Extracting Feature Code")
    logger.start_status("Starting Extraction")

    extracted = False
    entry_file = None
    extra_dependencies = []
    dep_to_path_mappings = {}

    _,chunks,_,import_deps = await parser.parse_code(file_name)

    for item in import_deps:
        element = import_deps[item]
        for depItem in element:
            if depItem["alias"] != None:
                deps = [depItem['alias']]
            elif "ext_deps" in depItem.keys() and len(depItem["ext_deps"]) > 0:
                deps = depItem['ext_deps']
            else:
                deps = [depItem['dep']]
            
            extra_dependencies.extend(deps)

            # create dep to path mapping
            for singleDep in deps:
                dep_to_path_mappings[singleDep] = item

    for chunk in chunks:
        words = chunk.split(".")
        if len(words) > 1:
            route_info = getRoutePathAndMethodsOrReturnNone(chunk,chunks)
            if route_info is not None:
                if route_info[0] == path:
                    entry_file = await handleExtraction(project_root,chunk,file_name)
                    extracted = True
                    break
    
    if not extracted:
        # Not found
        raise ValueError("Route specified was not found in ",file_name)

    logger.update_status("Indexing Feature files")    
    # sort and index extracted files here
    indexer = Indexer(
        "gpt-5",
        "openai",
        'OPENAI_API_KEY',
        os.environ.get("OPENAI_API_KEY","")
    )
    indexer.setup()
    indexer_ignore = ["package.json"]

    if entry_file is not None:
        indexer_ignore.append(entry_file)

    # step 1 : get all extracted files
    all_files = list_all_files("./output")
    for file in all_files:

        if os.path.basename(file) in indexer_ignore:
            continue

        # step 2 : find out what folder they are suppose to be in\
        logger.update_status(f"Indexing file : {os.path.basename(file)}")    
        res = indexer.index(file)
        folder = res['category']

        path_in_file = file.split("/")
        if folder not in path_in_file: # check if file is already in the right folder
            # step 3 : move them to correct folder and step 4 : update file refrence in other files 
            await moveFile(file,f"./output/{folder}")

    # package into zip
    home = str(Path.home())
    template_path = f"{home}/.sb_zip"
    output_dir = f"{project_root}/output"

    if not os.path.exists(template_path):
        os.makedirs(template_path,exist_ok=True)

    
    logger.stop_status()

    template_name = Prompt.ask("[bold green]Enter Template Name [/bold green]")
    # input("What do you want to name this template : ")

    if not template_name.startswith("speedbuild_"):
        template_name = f"speed_build_{template_name}"

    template_path = f"{template_path}/{template_name}"
    
    shutil.make_archive(template_path, 'zip', output_dir)

    clear_folder(output_dir)

    checkTemplateForVersionUpdate(template_path)

    # debug template code
    # TODO : only do this if template version changes
    # try:
    test_environment_path = f"{home}/.sb/environment/express"
    debugTemplate(f"{template_path}/lts.zip")

    # clear and delete node modules folder
    clear_folder(f"{test_environment_path}/node_modules")
    os.rmdir(output_dir)
    # delete package_lock.json
    os.remove(f"{test_environment_path}/package-lock.json")

    # overwrite zip file
    shutil.make_archive(f"{template_path}/lts", 'zip', test_environment_path)
    shutil.make_archive(f"{template_path}/extra", 'zip', test_environment_path)
    # except:
    #     pass

    logger.print_message(f"Extraction Complete : template saved with name `{template_name}`")


system_prompt = """
You are an expert software debugger tasked with analyzing errors and providing actionable debugging steps. You have access to three tools: reading files, updating files, and running shell commands.

## Your Role
- Analyze error messages, logs, and command outputs to identify root causes
- Provide step-by-step debugging instructions
- Focus on systematic problem-solving rather than guessing
- Prioritize the most likely solutions first

## Response Format
You MUST respond with valid JSON in exactly this format:

```json
{
    "action": "read_file | update_file | run_command | end",
    "description": "Clear explanation of what this step accomplishes",
    "file_path": "path/to/file or null",
    "command": "shell command or null", 
    "file_action": "add | replace | remove or null",
    "file_lines": "line numbers (e.g., '15' or '10-20') or null",
    "new_code": "code content or null"
}
```

## Action Types

### read_file
Use when you need to examine file contents to understand the codebase, configuration, or identify issues.
- Set `file_path` to the file you want to read, must be absolute paths
- All other fields should be null

### update_file  
Use when you need to modify, add, or remove files to fix issues.
- Set `file_path` to target file
- Set `file_action` to:
  - `"add"`: Append content to existing file
  - `"replace"`: Replace specific lines or entire file
  - `"remove"`: Delete the file
- Set `file_lines` for targeted changes (e.g., "15" for line 15, "10-20" for lines 10-20)
- Set `new_code` to the content to add/replace
- Leave `command` as null

### run_command
Use to execute commands for testing, building, installing dependencies, or gathering system information.
- Set `command` to the shell command
- All other fields should be null
- Prefer specific, safe commands over broad operations

### end
Use when the issue is resolved or no further debugging is possible.
- All fields except `action` and `description` should be null

## Debugging Strategy

1. **Understand the Error**: Read error messages carefully, identify error types (syntax, runtime, import, etc.)

2. **Gather Context**: 
   - Read relevant files to understand the codebase
   - Check configuration files, dependencies, environment
   - Look for related error patterns

3. **Form Hypotheses**: Based on the error and context, determine the most likely causes

4. **Test Systematically**: 
   - Start with the most likely fix
   - Make one change at a time
   - Verify each change works before proceeding

5. **Validate Solutions**: After making changes, run tests or the original failing command

## Best Practices

### Safety First
- Make minimal changes to fix specific issues
- Avoid broad system modifications
- Back up important files when making significant changes
- Use version control commands when available

### Efficient Debugging
- Read error messages completely - they often contain the exact solution
- Check common issues first: missing dependencies, typos, path problems
- Look for recent changes that might have caused the issue
- Consider environment differences (Python versions, OS, etc.)

### Code Quality
- Follow existing code style and patterns
- Add comments for complex fixes
- Ensure changes don't break other functionality
- Use proper error handling

## Common Error Patterns

### Python Errors
- `ModuleNotFoundError`: Install missing packages or fix import paths
- `SyntaxError`: Check for typos, missing brackets, incorrect indentation  
- `AttributeError`: Verify object types and available methods
- `FileNotFoundError`: Check file paths and existence



## Context Awareness
- Consider the programming language and framework being used
- Adapt debugging approach to the specific technology stack
- Pay attention to environment details (OS, Python version, etc.)
- Look for patterns in error frequency and timing

## When to End
- The original failing command now succeeds
- You've identified the issue but it requires human expertise (complex architectural changes, security decisions, etc.)
- You've exhausted reasonable debugging options
- The error indicates a fundamental environment or setup issue beyond your scope

## Example Responses

**Reading a configuration file:**
```json
{
    "action": "read_file",
    "description": "Reading configuration file to check database connection settings",
    "file_path": "/home/user/app/config/database.yml",
    "command": null,
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

**Fixing a syntax error:**
```json
{
    "action": "update_file", 
    "description": "Fixing syntax error: adding missing closing parenthesis on line 23",
    "file_path": "/home/user/app/src/main.py",
    "command": null,
    "file_action": "replace", 
    "file_lines": "23-25",
    "new_code": "result = calculate_total(price, tax)"
}
```

**Add code to file:**
```json
{
    "action": "update_file", 
    "description": "Initialize result variable with value 5000 on line 23",
    "file_path": "/home/user/app/src/main.py",
    "command": null,
    "file_action": "add", 
    "file_lines": "23",
    "new_code": "result = 5000"
}
```

**Installing missing dependency:**
```json
{
    "action": "run_command",
    "description": "Installing missing numpy package required by the application", 
    "file_path": null,
    "command": "pip install numpy",
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

**Successful completion:**
```json
{
    "action": "end",
    "description": "Issue resolved: missing dependency installed and import error fixed",
    "file_path": null,
    "command": null,
    "file_action": null, 
    "file_lines": null,
    "new_code": null
}
```

Remember: Be systematic, make one change at a time, and always explain your reasoning clearly.

ALSO NOTE VERY IMPORTANT : update, add, remove File action should always state the file_lines needed for the action !!!
"""


"""

You are an autonomous debugging agent for an Express.js project. 
You MUST use the provided functions to investigate and fix issues.

!!do not give manual steps to follow.!!

CRITICAL: You MUST use function calls, not give instructions.

AVAILABLE FUNCTIONS (use absolute paths):
- getFileContent(filePath, fileName) -> str
- updateFile(filePath, fileName, action, update=None, line=None)
- executeCommand(command, framework)

## **Function Definitions(python)**

def getFileContent(filePath: str, fileName: str) -> str:
    ```Return the content of a file
    
    Args:
        filePath: The directory path (use absolute paths)
        fileName: The file name
    ```

def updateFile(filePath: str, fileName: str, action: str, update: str = None, line: str = None):
    ```Updates a file at the specified path by performing the given action.
    
    Args:
        filePath (str): The directory path where the file is located
        fileName (str): The name of the file to update
        action (str): "add", "replace", "remove", or "create"
        update (str, optional): Content for "add", "replace", "create" actions
        line (str, optional): Line number (e.g. "10") or range (e.g. "10-15")
    ```

def executeCommand(command: str, framework: str):
    ```Executes a shell command
    
    Args:
        command (str): The shell command to execute
        framework (str): Should be "express" for this project
    ```

RULES
* NEVER give manual steps or instructions to follow.
* ALWAYS use the functions to investigate and fix issues yourself.
* Perform exactly ONE function call per turn.
* Explain your reasoning BEFORE the function call in **no more than two brief sentences**.
* When you call `updateFile`, you must explain why the minimal change is safe before calling and include a one-line summary of the intended change immediately after the call (plain text).
* Make minimal changes only.
* If you are confused and cannot proceed, ask a single concise clarification question (no instructions).

PROCESS
1. Read the error/stack trace provided.
2. State (in ≤2 sentences) what you will inspect or fix and why.
3. Make exactly ONE function call (as specified above).
4. Wait for the result and proceed in the next turn.

EXAMPLE
I will inspect the main server startup file to find why the app doesn't bind the port.

`getFileContent("/absolute/path/to/project", "server.js")`

"""

















"""
You are an expert Node JS developer debugging an Express project. 
You are given error logs, stack traces, or failing tests and must identify and fix bugs.

## **Rules**

* Perform exactly **ONE action per turn**
* Never touch `package.json` - ignore it completely
* Make minimal changes only
* Explain your reasoning before each action
* Do not repeat yourself or repeatedly call a tool without reason

NOTE!! Do not give me steps instead use the available tools to debug and fix the bug yourself 

## **Available Tools**

* `readFile` - Read a file
      Return the content of a file

      NOTE : always use the absolute path of the file !!!

      Args:
          filePath: first str
          fileName: second str

* `updateFile` - Update a file 
    Updates a file at the specified path by performing the given action.

    Args:
        filePath (str): The directory path where the file is located.
        fileName (str): The name of the file to update.
        action (str): The action to perform on the file. Accepted values are "add", "replace", "remove", or "create".
        update (str, optional): The content to add or replace in the file. Required for "add", "replace", and "create" actions.
        line (str, optional): The specific line number (not the actual line content) to update. Used for "replace" or "remove" actions.
    
    Example:
        //for single line action
        updateFile('./output','hello.js','replace',"console.log('hello')",'10')

        //for multi line action
        updateFile(
            './output',
            'hello.js',
            'replace',
            "console.log('hello'); console.log("world")",
            '10-12')
    
    Returns:
        str: An error message if the file is not found or the action is not recognized; otherwise, None.
    
    Notes:
        - For "add", "replace", and "remove" actions, the file must already exist.
        - For "create" action, a new file is created with the provided content.
        - The function relies on a helper function `performUpdate` to process the file content.


* `executeCommand` - Run a shell command
    Executes a given shell command using the PythonExecutor.
    framework (str) : express

    Args:
        command (str): The shell command to execute as a string.

    Behavior:
        - Splits the command string into a list of arguments.
        - Determines if the command should be self-exiting based on whether the commands start the server.
        - Executes the command in a specified working directory and environment.
        - Return the standard output and standard error of the executed command.

## **Process**

1. Choose ONE tool based on the error
2. Explain why you're using this tool
3. Use the tool
4. Wait for next turn

Stop when the issue is resolved or you need human help.

"""


import os
import sys
import venv
import shutil
from pathlib import Path

from .break_prompt import splitPrompt
from .src.Django.utils.deploy import read_yaml
from .src.Django.utils.venv_utils import get_activated_venv
from .utils.installDependencies import install_dependencies

from .move_django_to_target import moveFilesToTargetProject

from .langgraph_agent.agents.workflow.tools import updateFile

from .langgraph_agent.agents.workflow.claude_workflow import SpeedBuildWorkflow

from .exec.runCommand import PythonExecutor
from .exec.sb_exec_utils import create_virtual_environment,install_packages_in_venv
from .old_sb import implementFeature

def clear_folder(folder_path):
    for file_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, file_name)
        if os.path.isfile(file_path) or os.path.islink(file_path):  
            os.remove(file_path)  # Delete files & symlinks
        elif os.path.isdir(file_path):  
            shutil.rmtree(file_path)  # Delete directories


def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_paths.append(absolute_path)
    return file_paths

def getProjectFiles(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root or "__pycache__" in root or "migrations" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_name = absolute_path.replace(path,"").lstrip("/")
            if os.path.basename(file_name) not in ["__init__.py","apps.py",".speedbuild","manage.py"] and file_name.endswith("sqlite3")==False:
                file_paths.append(file_name)
    return file_paths


class SpeedBuildTestAgent():
    def __init__(self):
        self.home = str(Path.home())
        self.environment_sb_path = f"{self.home}/.sb/environment/"
        self.executor = PythonExecutor()
        venv_path = f"{self.home}/.sb/environment/django/venv"

        if sys.platform == "win32":
            python_path =  os.path.join(venv_path, "Scripts", "python.exe")
        else:
            python_path = os.path.join(venv_path, "bin", "python")

        # Modify environment to use the venv
        self.env = os.environ.copy()
        self.env['VIRTUAL_ENV'] = venv_path
        
        if sys.platform == "win32":
            self.env['PATH'] = f"{os.path.join(venv_path, 'Scripts')};{self.env['PATH']}"
        else:
            self.env['PATH'] = f"{os.path.join(venv_path, 'bin')}:{self.env['PATH']}"


    def installAndCreateDjangoProject(self,venv):
        res = create_virtual_environment(venv)
        installation_successful = install_packages_in_venv(f"{venv}/venv",["django"])
        # print("## virtual environment is ",self.env)
        if installation_successful:

            self.venv_path = f"{venv}/venv"
            print("### here")
            # create django project
            res = self.executor.runCommand(
                ["django-admin","startproject","speedbuild_project"],
                True,
                env=self.env,
                cwd="/home/attah/.sb/environment/django"
            )

            print(res)

            # create django app
            res = self.executor.runCommand(
                ["python","manage.py","startapp","speedbuild_app"],
                True,
                env=self.env,
                cwd="/home/attah/.sb/environment/django/speedbuild_project"
            )

            print(res)

            # add urls.py file
            shutil.copy("/home/attah/Documents/sb_final/speedbuild/exec/app_urls.py", "/home/attah/.sb/environment/django/speedbuild_project/speedbuild_app/urls.py")

            # include new urls file in main urls file
            shutil.copy("/home/attah/Documents/sb_final/speedbuild/exec/main_urls.py", "/home/attah/.sb/environment/django/speedbuild_project/speedbuild_project/urls.py")

            # do customization / test customization

            # move code to main project


    def createTestEnvironment(self,framework):
        environment_path = f"{self.environment_sb_path}{framework}"

         # clear out existing project
        print(environment_path, "hert ")
        clear_folder(environment_path)

        if not os.path.exists(environment_path):
            os.makedirs(environment_path,exist_ok=True)

        action = self.installAndCreateDjangoProject(environment_path)

        return environment_path+"/speedbuild_project"


    def testAndFixErrors(self):
        # run server -- check for error / test feature, troubleshoot and fix issues
        pass


    def deployFeatureInTestEnvironment(self,template_path,project_path):
        """Call speedbuild function to deploy code in test environment and also add speedbuild_app to installed apps"""

        implementFeature(template_path,project_path,"speedbuild_app",None,True,self.venv_path)
        print("Feature Implemented")


    def migrateTestAndFixErrors(self,max_debugger_count=3):
        commands = [
            "python manage.py makemigrations",
            "python manage.py migrate",
            "python manage.py runserver"
        ]

        llm_agent = SpeedBuildWorkflow("gpt-4o","openai","django")

        # if prompt == None: # normal debugging
        
        # Create workflow for debugging a specific command
        llm_agent.create_debugger_workflow()
            
        # # Run the initial failing command to get the first error
        # llm_agent.run(context)

        for command in commands:
           # Get initial error and command output
            no_errors,stdout,stderr = llm_agent.run_shell_command(command)

            if no_errors == False:
                llm_agent.last_error = stderr 
                llm_agent.last_command_output = stdout
                llm_agent.original_command = command

                llm_agent.run()

    def TestFeature(self,framework,template_path):
        project_path = self.createTestEnvironment(framework)

        # deploy template [add speedbuild_app to INSTALLED_APPS settings file]
        self.deployFeatureInTestEnvironment(template_path,project_path)

        # fix error
        self.migrateTestAndFixErrors()


    def customize_feature_code(self,prompt,files,root_path):
        llm_agent = SpeedBuildWorkflow("gpt-4o","openai","django")
        # if prompt == None: # normal debugging
        
        # Create workflow for debugging a specific command
        llm_agent.create_debugger_workflow("customization")

        llm_agent.last_command_output = f"""
        customization prompt : **{prompt}**
        project root path : **{root_path}***
        project files : {files}
        """

        llm_agent.run()

        if llm_agent.success:
            print("Customization successful")
            return True
        else:
            print("workflow ended without resolution")
            return False


def testCustomizeAndDeployDjangoFeature(feature_template,target_destination,target_app_name,customize=False):
    files = getProjectFiles(target_destination)
    target_project_name = None

    for file in files:
        if os.path.basename(file) == "settings.py":
            target_project_name = os.path.dirname(file)
    
    if target_project_name == None:
        raise ValueError("Could not detect project name, could not locate the settings.py file")
    
    
    test_environment =  f"{str(Path.home())}/.sb/environment/django/speedbuild_project"
    djangoAgent = SpeedBuildTestAgent()

    djangoAgent.TestFeature("django",feature_template)

    # ask for customization prompt
    if customize:
        prompt = input("Enter customization prompt, be as detailed as possible : \n")

        # TODO : break prompt into simple steps
        customize_prompts = splitPrompt(prompt)

        for prompt in customize_prompts:
            print("Processing",prompt)
            files = getProjectFiles(test_environment) 
            status = djangoAgent.customize_feature_code(prompt,files,test_environment)

        # TODO : if could not customize break out 

    # move feature code to target destination
    moveFilesToTargetProject(
        test_environment,
        target_destination,
        target_project_name,
        target_app_name
    )

    # install packages in target destination
    print("Installing Feature dependencies")
    config_name = feature_template.replace("speed_build_","")
    template_root = f"{str(Path.home())}/.sb_zip"
    venv = get_activated_venv()
    if ":" in feature_template:
        template_name, version = feature_template.split(":")

        yaml_path = f"{template_root}/{template_name.strip()}/{version.strip()}/sb_{config_name}.yaml"
    else:
        version = "lts"
        yaml_path = f"{template_root}/{feature_template}/{version}/sb_{config_name}.yaml"
    
    if not os.path.exists(yaml_path):
        raise ValueError(f"Template {template_name} version {version} does not exist")
    
    template_yaml = read_yaml(yaml_path)

    install_dependencies(template_yaml['dependencies'],venv)

if __name__ == "__main__":
    testCustomizeAndDeployDjangoFeature("speed_build_RegisterWithOauth","/home/attah/Documents/speedbuild/test_environment/mybookshop","shop",True)


import os
from pydantic import BaseModel,Field

from langchain.chat_models import init_chat_model
from langgraph.graph import START, END, StateGraph, MessagesState
from langchain_core.messages import AIMessage, SystemMessage, HumanMessage

from .indexJSSystemPrompt import system_prompt

class LLMOutput(BaseModel):
    category: str = Field(description="The category of the file")

class ExtendedMessagesState(MessagesState):
    # Add custom fields while keeping messages
    formatted_output: dict = {}
    structured_response: LLMOutput = None

class Indexer():
    def __init__(self,model,provider,key_name,api_key):
        # TODO : set model API key
        os.environ[key_name] = api_key
        
        llm = init_chat_model(model=model, model_provider=provider)
        self.llm = llm.with_structured_output(LLMOutput)

    def llm_call(self,state:ExtendedMessagesState):
        response = self.llm.invoke(state["messages"])

        return {
            "messages": [AIMessage(content=response.category)],
            "structured_response": response
        }

    def format_node(self,state: ExtendedMessagesState):
        response = state["structured_response"]
        formatted = {"category": response.category}
        return {"formatted_output": formatted}
    
    def setup(self):
        builder = StateGraph(ExtendedMessagesState)
        builder.add_node("llm",self.llm_call)
        builder.add_node('format_node',self.format_node)

        builder.add_edge(START, 'llm')
        builder.add_edge('llm','format_node')
        builder.add_edge('format_node',END)

        self.model = builder.compile()

    def initialize_with_system_message(self,user_input: str, system_prompt: str):
        return [
            SystemMessage(content=system_prompt),
            HumanMessage(content=user_input)
        ]
    
    def getFileInfo(self,filename):
        if not os.path.exists(filename):
            return None
        
        info = "file name : " + os.path.basename(filename) + "\n\n"

        # TODO : shorten contents of long files
        with open(filename,"r") as file:
            content = file.read()
            info += f"file content : ** {content} **"

        return info
    
    def index(self,filename):
        user_input = self.getFileInfo(filename)
        if user_input is None:
            return None
        
        messages = self.initialize_with_system_message(user_input,system_prompt)
        res =  self.model.invoke({"messages":messages})
        return res['formatted_output']

# if __name__ == "__main__":
#     indexer = Indexer("gpt-5","openai")
#     indexer.setup()
#     indexer.index("filename",system_prompt)



import json


data =  {
            'express': 
                [
                    {'dep': 'express', 'alias': "jumbo", 'use-require': True}
                ], 
            '../controllers/contact_controller.js': 
                [
                    {'dep': 'getContacts', 'alias': "contacts", 'use-require': True, 'standalone': True},
                    {'dep': 'createContact', 'alias': None, 'use-require': True, 'standalone': True}, 
                ], 
            '../middleware/auth_middleware.js': [
                {'dep': 'validateToken', 'alias': None, 'use-require': True}
                ]
        }


data2 =  {
            'express': 
                [
                    {'dep': 'express', 'alias': None, 'use-require': True}
                ], 
            '../controllers/contact_controller.js': 
                [
                    {'dep': 'updateContact', 'alias': None, 'use-require': True, 'standalone': True}, 
                    {'dep': 'deleteContact', 'alias': None, 'use-require': True, 'standalone': True}
                ], 
            '../middleware/auth_middleware.js': [
                {'dep': 'validateToken', 'alias': None, 'use-require': True}
                ]
        }


def mergeImportDict(dict1,dict2):
    for item in dict2:
        if item in dict1.keys():
            for entry in dict2[item]:
                if entry not in dict1[item]:
                    dict1[item].append(entry)
        else:
            dict1[item] = dict2[item]
            
    return dict1

def formES6Import(nameImport,others,dep):
    import_str = "import "
    if len(nameImport) > 0:
        import_str += "{" + ','.join(nameImport) + "} "

    if len(others) > 0:
        if import_str.lstrip().endswith("}"):
            import_str += ","
        import_str += f"{','.join(others)} "

    import_str += f'from "{dep}"'
    return import_str


def formRequireStyleImport(namedImport,others,dep):
    import_str = ""
    if len(namedImport) > 0:
        import_str += "{" + ','.join(namedImport) + "} "

    if len(others) > 0:
        if import_str.lstrip().endswith("}"):
            import_str += ","
        import_str += f"{','.join(others)} "

    import_str = f'const {import_str} = require("{dep}")'
    return import_str


def convertToImportStatements(data):
    # TODO we've not added alias or accounted for import with aliases
    imports = []

    for i in data:
        element = data[i]
        nameInport = []
        others = []
        use_require = element[0]["use-require"]


        for item in element:
            newImport = f"{item['dep']} as {item['alias']}" if item['alias'] is not None else item['dep']

            if "standalone" not in item.keys():
                others.append(newImport)
                continue

            if item['standalone'] == False:
                nameInport.append(newImport)
            else:
                others.append(newImport)

        if use_require:
            import_str = formRequireStyleImport(nameInport,others,i)
        else:
            import_str = formES6Import(nameInport,others,i)

        imports.append(import_str)

    return imports

def getDepFromImportDict(data,dep_names):
    deps = []

    for entry in data:
        if entry['dep'].strip() in dep_names:
            deps.append(entry)

    if len(deps) > 0:
        return deps
    
    return None

# print(json.dumps(mergeImportDict(data,data2),indent=4))
# print(convertToImportStatements(data))


import io
import os
import zipfile
import requests

from pathlib import Path

from .server_address import url_address

def pullInitialTemplates():
    output_folder = str(Path.home()) + "/.sb_zip"
    os.makedirs(output_folder, exist_ok=True)

    try:
        res = requests.get(url_address + "sb/initial-templates/", stream=True, timeout=10)
        res.raise_for_status()
    except requests.exceptions.RequestException as e:
        print(f"Network error occurred: {e}")
        return

    file_bytes = io.BytesIO(res.content)
    try:
        with zipfile.ZipFile(file_bytes) as z:
            z.extractall(output_folder)
    except zipfile.BadZipFile:
        pass
        print("Downloaded file is not a valid zip archive.")


system_prompt = """
# SpeedBuild AI Agent - System Prompt

You are a **smart AI agent** specialized in **Express.js** and experienced in **production-grade Node.js code modification**. You are part of SpeedBuild, a service that helps developers extract, customize, and reuse Express.js features using plain English commands.

## Core Mission
Extract and customize Express.js features while maintaining production reliability, performance, and Node.js best practices.

---

## Available Tools

### 1. read_file
**When to use**: Examine file contents, understand codebase structure, or identify issues
- `file_path`: Absolute path to target file
- All other fields: `null`

### 2. update_file  
**When to use**: Modify, add, or remove files
- `file_path`: Absolute path to target file
- `file_action`: 
  - `"add"`: Insert new content at specified line
  - `"replace"`: Replace content at specified lines
  - `"remove"`: Delete content at specified lines
- `file_lines`: Target line(s) - single number ("15") or range ("10-20")
- `new_code`: Content to add/replace (not needed for remove)
- `command`: Always `null`

### 3. run_command
**When to use**: Execute shell commands for testing, building, or installing dependencies
- `command`: Shell command to execute
- All other fields: `null`

### 4. end
**When to use**: Task completed or no further action possible
- `description`: Summary of completion status
- All other fields: `null`

---

## Workflow Strategy

### 1. Understand Request
- Analyze the customization goal and scope
- Identify affected Express.js components (routes, middleware, controllers, models, services)
- Determine file processing order and dependencies

### 2. Gather Context
- Read relevant files to understand current implementation
- Check configurations, dependencies, and environment setup
- Identify potential breaking changes or dependencies

### 3. Execute Changes
- **Process dependencies first** before main feature modifications
- Make **one change at a time** and verify impact
- Follow Express.js patterns and conventions
- Add meaningful comments for complex changes

### 4. Validate Results
- Run tests or commands to verify functionality
- Check for broken imports or syntax errors
- Ensure no regression in existing features

---

## Express.js Best Practices

### Routes
- Preserve authentication/authorization middleware
- Maintain error handling patterns
- Keep request/response validation
- Use proper HTTP status codes
- Maintain route parameter validation

### Middleware
- Preserve middleware order and dependencies
- Keep error handling middleware at the end
- Maintain authentication flows
- Preserve CORS and body parsing configurations

### Controllers
- Maintain async/await patterns
- Preserve input validation and sanitization
- Keep consistent response formats
- Preserve database transaction patterns

### Models/Schemas
- Maintain schema validation (Mongoose)
- Preserve model relationships
- Keep indexing optimizations
- Preserve schema middleware (hooks)

### Services
- Keep business logic encapsulation
- Maintain error handling patterns
- Preserve external integrations
- Keep caching patterns

---

IMPORTANT NOTES
1) always re-read file after every update to confirm changes
2) updates are done one at a time.
3) confirm if customization was successful

---

## Response Format

**REQUIRED**: All responses must be valid JSON with this exact structure:

```json
{
    "action": "read_file | update_file | run_command | end",
    "description": "Clear explanation of this action",
    "file_path": "absolute/path/to/file or null",
    "command": "shell command or null", 
    "file_action": "add | replace | remove or null",
    "file_lines": "line number(s) or null",
    "new_code": "code content or null"
}
```

### File Action Requirements
- **add**: Requires `file_lines` (insertion point) and `new_code`
- **replace**: Requires `file_lines` (target range) and `new_code`
- **remove**: Requires only `file_lines` (deletion range)

---

## Safety Guidelines

### Code Modifications
- Make **minimal, targeted changes**
- Preserve existing functionality and patterns
- Add comments explaining complex modifications
- Maintain proper indentation and formatting
- Follow existing code style conventions

### File Paths
- Always use **absolute paths** (e.g., `/home/user/app/routes/users.js`)
- Never use relative paths or shortcuts

### Error Handling
- If unable to safely modify, use `"end"` action with explanation
- Test changes before proceeding to next modification
- Preserve existing error handling patterns

---

## Example Responses

### Reading a file
```json
{
    "action": "read_file",
    "description": "Reading user routes to understand current authentication flow",
    "file_path": "/home/user/app/routes/users.js",
    "command": null,
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

### Adding new code
```json
{
    "action": "update_file",
    "description": "Adding input validation middleware for user registration",
    "file_path": "/home/user/app/routes/users.js",
    "command": null,
    "file_action": "add",
    "file_lines": "15",
    "new_code": "// Validate user registration input\nconst validateUser = require('../middleware/validateUser');\nrouter.use('/register', validateUser);"
}
```

### Replacing code
```json
{
    "action": "update_file",
    "description": "Updating error response format to include error codes",
    "file_path": "/home/user/app/controllers/userController.js",
    "command": null,
    "file_action": "replace",
    "file_lines": "45-47",
    "new_code": "return res.status(400).json({\n  success: false,\n  error: {\n    code: 'VALIDATION_ERROR',\n    message: error.message\n  }\n});"
}
```

### Running command
```json
{
    "action": "run_command",
    "description": "Installing bcrypt package for password hashing",
    "file_path": null,
    "command": "npm install bcrypt",
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

### Task completion
```json
{
    "action": "end",
    "description": "Successfully added password hashing to user registration with input validation",
    "file_path": null,
    "command": null,
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

---

## Completion Checklist

Before using `"end"` action, verify:
- All requested customizations implemented
- No broken imports or syntax errors
- Express.js patterns maintained
- Core functionality preserved
- All relevant files reviewed
- Dependencies properly handled
- API contracts remain consistent

---

## Key Reminders

1. **One change at a time** - Make incremental, testable modifications
2. **Absolute paths only** - Never use relative file paths
3. **Preserve patterns** - Maintain Express.js conventions and existing code style  
4. **Add comments** - Explain complex changes for maintainability
5. **Safety first** - When in doubt, ask for clarification rather than guess
6. **Test changes** - Verify functionality after each modification
"""


















"""
```
You are a **smart AI agent** proficient in **Express.js** and experienced in **production-grade Node.js code modification**. 
You are part of a service called **SpeedBuild**, which helps developers extract features from their code, 
customize those features using plain English commands, and reuse them in new projects.
You have access to three tools: reading files, updating files, and running shell commands.

## Action Types

### read_file
Use when you need to examine file contents to understand the codebase, configuration, or identify issues.
- Set `file_path` to the file you want to read, must be absolute paths
- All other fields should be null

### update_file  
Use when you need to modify, add, or remove files to fix issues.
- Set `file_path` to target file
- Set `file_action` to:
  - `"add"`: Append content to existing file
  - `"replace"`: Replace specific lines or entire file
  - `"remove"`: Delete the file
- Set `file_lines` for targeted changes (e.g., "15" for line 15, "10-20" for lines 10-20)
- Set `new_code` to the content to add/replace
- Leave `command` as null

### run_command
Use to execute commands for testing, building, installing dependencies, or gathering system information.
- Set `command` to the shell command
- All other fields should be null
- Prefer specific, safe commands over broad operations

### end
Use when the issue is resolved or no further debugging is possible.
- All fields except `action` and `description` should be null

## Debugging and Customization Strategy

1. **Understand the user request**

2. **Gather Context**: 
   - Read relevant files to understand the codebase
   - Check configuration files, dependencies, environment
   - Look for related error patterns

3. **Form Hypotheses**: Based on user prompt request and context, determine the actions to take

4. **Test Systematically**: 
   - Start with the most likely action
   - Make one change at a time
   - Verify each change works before proceeding

5. **Validate Solutions**: After making changes, run tests to verify code is working and bugs were not introduced

## Best Practices

### Safety First
- Make minimal changes to fix specific issues
- Avoid broad system modifications

### Efficient Debugging
- Read error messages completely - they often contain the exact solution
- Check common issues first: missing dependencies, typos, path problems
- Look for recent changes that might have caused the issue
- Consider environment differences (Python versions, OS, etc.)

### Code Quality
- Follow existing code style and patterns
- Add comments for complex fixes
- Ensure changes don't break other functionality
- Use proper error handling


## Response Format
You MUST respond with valid JSON in exactly this format:

```json
{
    "action": "read_file | update_file | run_command | end",
    "description": "Clear explanation of what this step accomplishes",
    "file_path": "path/to/file or null",
    "command": "shell command or null", 
    "file_action": "add | replace | remove or null",
    "file_lines": "line numbers (e.g., '15' or '10-20') or null",
    "new_code": "code content or null"
}
```


## Your Core Responsibilities

A feature and its dependencies may span multiple files. Your job is to:

1. **Analyze the user's customization request** to understand their intent and scope
2. **Process dependencies first** before modifying the main feature
3. **Retrieve file contents when needed** by requesting the user to provide them
4. **Modify code while maintaining Express.js best practices and structure**
5. **Ensure all modifications include meaningful comments** to aid understanding
6. **Return responses in structured JSON format** for automated processing
7. **Review every file listed by the user** to determine if modifications are needed
8. **Preserve the feature's core reliability and functionality**
9. **Handle Express-specific patterns** (routes → middleware → controllers → models → services)
10. **Terminate only when all files are reviewed and the customization is complete**

---

## User Intent Analysis

Before starting modifications, analyze the customization request to understand:

* **Primary Goal**: What is the user trying to achieve?
* **Scope**: Which Express components will be affected (routes, middleware, controllers, models, services)?
* **Constraints**: What should remain unchanged to preserve functionality?
* **Dependencies**: What order should files be processed in?
* **Express Patterns**: Will this require new routes, middleware updates, or database schema changes?

---

## Express.js-Specific Intelligence

When working with Express.js code, maintain these patterns:

### Route Modifications
* Preserve authentication and authorization middleware
* Maintain consistent error handling patterns
* Keep request/response validation intact
* Ensure proper HTTP status codes and response formats
* Maintain route parameter validation

### Middleware Changes
* Preserve middleware order and dependencies
* Keep error handling middleware at the end
* Maintain authentication flows
* Preserve logging and security middleware
* Keep CORS and body parsing configurations

### Controller Updates
* Maintain async/await patterns and error handling
* Preserve input validation and sanitization
* Keep business logic separation
* Maintain consistent response formats (JSON APIs)
* Preserve database transaction patterns

### Model/Schema Changes
* If using Mongoose, maintain schema validation
* Preserve model relationships and references
* Keep indexing and performance optimizations
* Maintain schema middleware (pre/post hooks)
* Preserve custom model methods

### Service Layer
* Keep business logic encapsulation
* Maintain error handling patterns
* Preserve external API integrations
* Keep database abstraction layers
* Maintain caching patterns

### Configuration Files
* Preserve environment variable usage
* Maintain security configurations
* Keep database connection settings
* Preserve logging configurations
* Maintain package.json dependencies

---

## Structured JSON Response Format

Always return a structured JSON response with these top-level fields:

### Action Types
* **`"read file"`** → When you need the content of a file
* **`"update file"`** → when you want to update or create a new file
* **`"run command"`** → When you want to run a shell command
* **`"end"`** → When you want to terminate the customization workflow

### Response Fields
* **`action`**: One of the action types above
* **`description`**: A short description of the action being taken
* **`file_path`**: The path of the file being referenced or changed
* **`command`**: The shell command to be executed
* **`file_action`**: The type of action to take on file : add, replace or remove
* **`file_lines`**: used along file_action hold the line or line range of code to replace, remove or line to insert new code
* **`new_code`**: Code to add to file

NOTE!!! `file_lines` can be a single number e.g 10 or a range of number e.g 20 - 40

---

## Code Block Update Format

### Rules for update file
* **add**: both `new_code` and `file_lines` are required
* **Replace**: Both `new_code` and `file_lines` of old code to replace
* **Remove**: Only `file_lines` of code to remove is required
* Always include meaningful comments explaining the change
* Preserve original code formatting and style
* Maintain proper indentation and JavaScript/TypeScript conventions
* Keep consistent import/require patterns

---

## Error Handling and Safety

### When You Cannot Safely Modify Code
```json
{
  "action": "end",
  "description": "Cannot safely modify without breaking functionality : This middleware contains complex authentication logic that requires human review",
  "file_path": "/home/user/app/complex_middleware.js", 
}
```

---

## Context Preservation Rules

Throughout the customization session:

1. **Remember the overall goal** - each modification should work toward the user's stated objective
2. **Maintain feature integrity** - preserve the core purpose and reliability of the original feature
3. **Track dependencies** - ensure that changes in one file don't break imports in others
4. **Preserve Express patterns** - maintain the framework's conventions and best practices
5. **Consider production impact** - warn about changes that could affect existing APIs or data
6. **Maintain async patterns** - preserve Promise/async-await patterns and error handling

---

## Session Completion Checklist

Before setting action to `"end"`, verify:

* All requested customizations have been addressed
* No broken imports or require statements remain  
* Express.js-specific patterns are maintained
* The feature's core functionality is preserved
* All necessary files have been reviewed
* Any required follow-up actions are noted (npm installs, env vars, etc.)
* Middleware order and dependencies are intact
* API contracts remain consistent

---

## File Naming Convention

**When creating new files or requesting, always use absolute path.**

Examples:
* `/home/user/app/routes/api/users.js` (correct)
* `/home/user/app/middleware/auth.js` (correct)
* `/home/user/app/models/User.js` (correct)
* `/home/user/app/controllers/userController.js` (correct)
* `/home/user/app/app.js` (correct - root folder)

---

## Example Responses

**Reading a configuration file:**
```json
{
    "action": "read_file",
    "description": "Reading configuration file to check database connection settings",
    "file_path": "/home/user/app/config/database.yml",
    "command": null,
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

**Fixing a syntax error:**
```json
{
    "action": "update_file", 
    "description": "Fixing syntax error: adding missing closing parenthesis on line 23",
    "file_path": "/home/user/app/src/main.py",
    "command": null,
    "file_action": "replace", 
    "file_lines": "23-25",
    "new_code": "result = calculate_total(price, tax)"
}
```

**Add code to file:**
```json
{
    "action": "update_file", 
    "description": "Initialize result variable with value 5000 on line 23",
    "file_path": "/home/user/app/src/main.py",
    "command": null,
    "file_action": "add", 
    "file_lines": "23",
    "new_code": "result = 5000"
}
```

**Installing missing dependency:**
```json
{
    "action": "run_command",
    "description": "Installing missing numpy package required by the application", 
    "file_path": null,
    "command": "pip install numpy",
    "file_action": null,
    "file_lines": null,
    "new_code": null
}
```

**Successful completion:**
```json
{
    "action": "end",
    "description": "Issue resolved: missing dependency installed and import error fixed",
    "file_path": null,
    "command": null,
    "file_action": null, 
    "file_lines": null,
    "new_code": null
}
```

Remember: Be systematic, make one change at a time, and always explain your reasoning clearly.

ALSO NOTE VERY IMPORTANT : update, add, remove File action should always state the file_lines needed for the action !!!

Remember again, `Your goal is to safely customize proven, production-ready Express.js features while maintaining their reliability, performance, and Node.js best practices`.
```
"""


import ast

def removeDecorator(code):
    if code == None:
        return ""
    blocks = code.split("\n")
    index = None

    for i in range(len(blocks)):
        if blocks[i].startswith("@"):
            index = i
        else:
            break
    
    if index != None:
        code_without_decorator = blocks[index:]
        return "\n".join(code_without_decorator).strip()
    
    return "\n".join(blocks).strip()


def get_assigned_variables(block,root=False):
    """
    Extracts variable names assigned within the given block of Python code using the Abstract Syntax Tree (AST).
    Args:
        block (str): The block of Python code to analyze.
        root (bool, optional): If True, returns only the first assigned variable found. Defaults to False.
    Returns:
        set or str: A set of assigned variable names if root is False, or a single variable name (str) if root is True.
    Notes:
        - Handles direct assignments, type-annotated assignments, function and class definitions.
        - Ignores decorators by removing them before parsing.
        - Returns an empty set if parsing fails or no assignments are found.
    """
    """Extracts variable names assigned within the given block using AST."""
    assigned_vars = set()
    
    block = removeDecorator(block)
    # print("hee  ", block, "\n\n")
    try:
        tree = ast.parse(block.strip())
        
        for node in ast.walk(tree):
            if isinstance(node, ast.Assign):  # Direct assignments (x = 5, x, y = 10, 20)
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        assigned_vars.add(target.id)
            elif isinstance(node, ast.AnnAssign):  # Type-annotated assignments (x: int = 5)
                if isinstance(node.target, ast.Name):
                    assigned_vars.add(node.target.id)
            elif isinstance(node, ast.FunctionDef):  # Function parameters
                assigned_vars.add(node.name)
                # for arg in node.args.args:
                #     assigned_vars.add(arg.arg)
            elif isinstance(node, ast.ClassDef):  # Exclude class names from assigned vars
                assigned_vars.add(node.name)

            if root and len(assigned_vars) > 0:
                return list(assigned_vars)[0]
    except:
        pass
        # print("error for ",block)

    return assigned_vars


from rich.console import Console

class StatusManager:
    def __init__(self):
        self.console = Console()
        self.status = None

    def start_status(self,msg):
        self.status = self.console.status(f"[bold cyan]{msg}[/bold cyan]")
        self.status.start()
        return self.status

    def update_status(self, message):
        if self.status:
            self.status.update(f"[bold cyan]{message}[/bold cyan]")

    def stop_status(self, msg=None):
        if self.status:
            self.status.stop()
            if msg:
                self.console.print(f"[bold cyan]{msg}[/bold cyan]")

    def print_message(self,msg):
        self.console.print(f"[cyan]{msg}[/cyan]")


import os
import ast
import requests
from pathlib import Path
from typing import List, Tuple

from .server_address import url_address
from ..src.Django.utils.django_app_dependencies import save_cache


def getCurrentDjangoFiles(path):
    python_files = [file for file in os.listdir(path) if file.endswith(".py")]
    return python_files

def pullPythonPackageJSON():
    try:
        res = requests.get(url_address + "sb/packages/", timeout=10)
        res.raise_for_status()
        print(res.json())

        save_cache(res.json())

    except requests.exceptions.HTTPError as http_err:
        raise ValueError(f"HTTP error occurred: {http_err} - Status code: {res.status_code}")
    except requests.exceptions.ConnectionError:
        raise ValueError("Connection error. Is the server running?")
    except requests.exceptions.Timeout:
        raise ValueError("Request timed out.")
    except Exception as err:
        raise ValueError (f"An error occurred: {err}")


def extract_views_from_urls(urls_file_path: str) -> List[Tuple[str, str]]:
    print(f"Extracting views from {urls_file_path}")
    with open(urls_file_path, "r") as f:
        data = f.read()
        data = data.replace(".as_view()", "")

        tree = ast.parse(data, filename=urls_file_path)

        view_names = []
        import_alias_map = {}

        # Step 1: Capture import statements to trace views
        for node in ast.walk(tree):
            if isinstance(node, ast.ImportFrom):
                module = node.module
                if module == None:
                    module = ""
                for alias in node.names:
                    import_alias_map[alias.asname or alias.name] = f"{module}.{alias.name}"
            elif isinstance(node, ast.Import):
                for alias in node.names:
                    import_alias_map[alias.asname or alias.name] = alias.name

        # Step 2: Find calls to `path()` or `re_path()`
        for node in ast.walk(tree):
            
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id in ("path", "re_path"):
                
                if len(node.args) >= 2:
                    view = node.args[1]
                    if isinstance(view, ast.Attribute):

                        # Something like views.HomeView
                        value_id = view.value.id if isinstance(view.value, ast.Name) else None

                        attr = view.attr
                        full_path = import_alias_map.get(value_id)
                        if full_path:
                            view_names.append(f"{full_path}.{attr}")
                    elif isinstance(view, ast.Name):
                        # Direct function/class reference
                        view_path = import_alias_map.get(view.id)
                        if view_path:
                            view_names.append(view_path)

    return view_names


def findFilePath(path,filename):
    # print("searching ",path)
    file_paths = []
    for root, _, files in os.walk(path):
        for file in files:
            # Get the relative path to the base directory
            if filename in file:
                relative_path = os.path.relpath(os.path.join(root, file), path)
                file_paths.append(relative_path)

    return file_paths


def getAbsolutePath(relative_path):
    relative_path = Path(relative_path)
    absolute_path = relative_path.resolve()

    return str(absolute_path)

def get_template_output_path(feature_name):
    user_dir = str(Path.home())
    user_dir += "/.sb_zip"

    if not os.path.exists(user_dir):
        os.makedirs(user_dir)

    output_zip = f"{user_dir}/speed_build_{feature_name}"  # No .zip extension needed

    return output_zip


import re


# Extract imports
def extract_imports(lines):
    return [line for line in lines if line.startswith("from") or line.startswith("import")]

# Extract urlpatterns
def extract_urlpatterns(lines):
    in_urlpatterns = False
    urls = []
    for line in lines:
        if "urlpatterns" in line:
            in_urlpatterns = True
            continue

        if line.strip() == "]":
            break

        elif in_urlpatterns:
            if len(line.strip()) > 0:
                urls.append(line)

    return urls

# Extract router definitions and rename if necessary
def extract_router(lines, suffix):
    router_defs = {}
    router_lines = []
    router_pattern = re.compile(r"^(\w+)\s*=\s*(\w+Router)\(\)")  # Matches `router = RouterType()`

    for line in lines:
        match = router_pattern.match(line)
        if match:
            router_name, router_type = match.groups()
            new_router_name = f"{router_name}_{suffix}"  # Rename to avoid conflicts
            line = line.replace(router_name, new_router_name)
            router_defs[new_router_name] = router_type
            router_lines.append(line)

    return router_lines, router_defs

# Extract direct `urlpatterns += router.urls` usage and rename routers
def extract_router_urlpatterns(lines, router_dict):

    direct_router_urls = []
    register_url = []

    for line in lines:
        if not(line.startswith("import ") or line.startswith("from ")):
            for old_name, _ in router_dict.items():
                new_name = old_name
                old_name = old_name.split("_")
                old_name.pop()
                old_name = "_".join(old_name)

                if old_name in line:
                    if "urlpatterns" in line:
                        delimeter = "="
                        if "+=" in line:
                            delimeter = "+="

                        code = line.split(delimeter)[1].strip().replace(old_name,new_name)

                        register_url.append(code)
                    else:
                        try:
                            if line.split("=")[1].strip().startswith(_):
                                continue
                        except ValueError:
                            pass

                        direct_router_urls.append(line.replace(old_name,new_name))

    return [direct_router_urls,register_url]


def merge_urls(urls1,urls2):

    imports1 = extract_imports(urls1)
    imports2 = extract_imports(urls2)

    # Merge imports without duplicates
    merged_imports = sorted(set(imports1 + imports2))

    urlpatterns1 = extract_urlpatterns(urls1)
    urlpatterns2 = extract_urlpatterns(urls2)

    merge_patterns = sorted(set(urlpatterns1+urlpatterns2))

    # Merge urlpatterns while ensuring all unique paths are included
    patterns = ["urlpatterns = [","\n"]
    patterns.extend(merge_patterns)
    patterns.extend("]")

    router_lines1, router_defs1 = extract_router(urls1, "1")
    router_lines2, router_defs2 = extract_router(urls2, "2")

    # Merge routers while preserving different router names
    merged_router_code = sorted(set(router_lines1 + router_lines2))

    router_urlpatterns1, register_rounter1_urls = extract_router_urlpatterns(urls1, router_defs1)
    router_urlpatterns2, register_rounter2_urls = extract_router_urlpatterns(urls2, router_defs2)

    # Merge router urlpatterns without duplication
    merged_router_urlpatterns = sorted(set(router_urlpatterns1 + router_urlpatterns2))
    merged_router_code += merged_router_urlpatterns

    # merge register_router_urls
    merge_register_router = sorted(set(register_rounter1_urls + register_rounter2_urls))

    # Ensure `include(router.urls)` is only added once
    if any("include(" in line for line in patterns) and merged_router_urlpatterns:
        merged_router_urlpatterns = []  # Avoid duplicate API URL inclusion


    url_code = merged_imports + ["\n"] + merged_router_code + ["\n\n"]

    url_code = [i for i in url_code if len(i.strip()) > 0]

    code = "\n".join(url_code) + "\n\n\n"

    url_code = patterns + [f" + {route}" for route in merge_register_router]
    url_code = [i for i in url_code if len(i.strip()) > 0]
    code += "\n".join(url_code)

    return code


stringIdentifier = ["'",'"',"`"]
route_identifies = ["route","get","post","put","delete","use"] # and others

def getFirstParameter(code):
    """
    Extracts the first parameter from a parenthesized expression in JavaScript/TypeScript code.
    The function processes code starting with an opening parenthesis and extracts characters
    until it encounters either a comma or matching closing parenthesis. It handles nested
    parentheses, brackets and braces properly.
    Args:
        code (str): A string starting with an opening parenthesis containing JavaScript/TypeScript code.
    Returns:
        str: The extracted first parameter as a string with whitespace preserved.
    Raises:
        ValueError: If the parentheses/brackets/braces in the code are not properly matched.
    Example:
        >>> getFirstParameter("(req, res)")
        'req'
        >>> getFirstParameter("({x: 1}, y)")
        '{x: 1}'
    """
    og = code
    opening = ["(","[","{"]
    close = [")","]","}"]
    stack = ["("] # add the starting parantesis to stack
    code = code[1:] # ignore the leading opening paranthesis
    parameter = ""

    for char in code:
        if len(stack) == 0:
            break 

        if char in opening:
            stack.append(char)

        elif char in close:
            if len(stack) > 0:
                last_in_stack = stack[-1]
                if last_in_stack in opening and opening.index(last_in_stack) == close.index(char):
                    stack.pop(-1) # remove last character in stack

        elif char == "," and len(stack) == 1:
            stack.pop() # remove opening bracket from stack
            break

        else:
            parameter += char

    if len(stack) > 0:
        print(stack)
        raise ValueError(f"You have a synthax error in your code : {og} ")
    
    return parameter

def manageIdentifier(indentifier, code):
    code = code.split(f".{indentifier}")
    if len(code) == 0:
        return None
    
    code = code[1]

    try:
        return getFirstParameter(code)
    except ValueError as error:
        print("Error : ",error)
        return None

def getRoutePathAndMethodsOrReturnNone(code,chunks):
    """
    Extracts route path and HTTP methods from Express.js-like route definitions.

    This function analyzes code chunks to identify route paths and their associated HTTP methods
    in Express.js-style route definitions.

    Args:
        code (str): The line of code to analyze
        chunks (list): Additional code chunks (currently unused)

    Returns:
        list or None: A list containing [path_name, methods_list] if a valid route is found,
                     where path_name is the cleaned route path (str) and methods_list contains
                     the HTTP methods (list of str). Returns None if no valid route is found
                     or if the line starts with '/'.

    Example:
        >>> code = "app.get('/users', handler)"
        >>> getRoutePathAndMethodsOrReturnNone(code, [])
        ['/users', ['get']]
    """
    pathName = None
    methods = []

    if code.startswith("/"):
        return None #Dont process comments

    for indentifier in route_identifies:
        if f".{indentifier}(" in code:
            if indentifier != "route":
                methods.append(indentifier)

            path = manageIdentifier(indentifier,code)
            if path is not None:
                if path[0] in stringIdentifier and path[-1] in stringIdentifier:
                    pathName = path

    if pathName is not None:
        # print(f"Path : {pathName} , methods : {methods} ",code )
        return [pathName.replace("'","").replace('"',""),methods] #TODO : change later
    

# code = """
# passport.use(
#   new GoogleStrategy(
#     {
#       clientID: process.env.GOOGLE_CLIENT_ID,
#       clientSecret: process.env.GOOGLE_CLIENT_SECRET,
#       callbackURL: process.env.GOOGLE_CALLBACK_URL,
#     },
#     async (accessToken, refreshToken, profile, done) => {
#       try {
#         // Find user by Google ID OR email
#         let user = await User.findOne({ email: profile.emails[0].value });

#         if (!user) {
#           user = await User.create({
#             name: profile.displayName,
#             email: profile.emails[0].value,
#             password: Math.random().toString(36).slice(-8), // dummy password
#             role: "user",
#           });

#           sendWelcomeEmail(user.name, user.email);
#         }

#         return done(null, user);
#       } catch (err) {
#         return done(err, null);
#       }
#     }
#   )
# );
# """

# code = """router.post("/register", async (req, res) => {
#     const { name, email, password, role } = req.body;
#     try {
#         const exists = await User.findOne({ email });
#         if (exists) return res.status(400).json({ message: "User already exist" });

#         const user = await User.create({ name, email, password, role });
#         const token = generateToken(user._id);
#         res.cookie("token", token, cookieOptions);

#         sendWelcomeEmail(user.name, user.email);
#         res.status(201).json({
#             _id: user.id,
#             name: user.name,
#             email: user.email,
#             role: user.role,
#             isAdmin: user.isAdmin,
#         });
#     } catch (err) {
#         res.status(500).json({ message: err.message});
#     }
# })"""

# data = getRoutePathAndMethodsOrReturnNone(code,[])
# print(data)



import asyncio
import os
from ..parser.jsParser import JsTxParser
from .jsDep import getChunkDependencies
from .extractJs import extractFeatureCode, getExtraDependencies
from .jsPath import getRoutePathAndMethodsOrReturnNone


parser = JsTxParser()

# TODO : extract external file dependencies

async def processEntryFile(file_name,file_dependencies,project_root):

    _,chunks,chunks_name,import_dict = await parser.parse_code(file_name)
    packages = []

    for chunk in chunks:
        deps = None
        skip_chunk = False
        words = chunk.split(".")
        if len(words) > 1:
            route_info = getRoutePathAndMethodsOrReturnNone(chunk,chunks)

            if route_info is not None:
                deps = getChunkDependencies(chunk,chunks_name,False)
                _, dep_to_path_mappings = getExtraDependencies(import_dict)

                for dep in deps:
                    if dep in dep_to_path_mappings.keys():
                        full_path = os.path.normpath(f"{project_root}/{dep_to_path_mappings[dep]}")
                        print("Full path is ", full_path)

                        # TODO : test this out
                        path_without_dot_or_slash = full_path.replace(project_root,"").lstrip("/")
                        print("Full path cleaned is ", path_without_dot_or_slash)
                        
                        if path_without_dot_or_slash not in file_dependencies:
                            skip_chunk = True
                            break

        # is_ignore = len([i for i in deps if dep_to_path_mappings[i] in file_dependencies]) > 0
        if not skip_chunk:
            print("Processing \n",chunk, "\n","#"*10)
            packages = await extractFeatureCode(chunk,file_name,False,packages,file_dependencies) # extract code

    print("Packages are ", packages)
    
if __name__ == "__main__":
    file_name = "express.js_contact_app/index.js"
    file_dependencies = ["./routes/contact_routes.js"]
    project_root = "express.js_contact_app"
    asyncio.run(processEntryFile(file_name, file_dependencies,project_root))


import os
import sys
from pathlib import Path

from ...newExec import PythonExecutor
# from ...exec.runCommand import PythonExecutor


# Read File
def getFileContent(filePath : str,fileName : str)->str:
    """Return the content of a file

    Args:
        filePath: first str
        fileName: second str
    """
    dest = f"{filePath}/{fileName}"
    
    # remove this
    if not dest.startswith("/home/attah/Documents/speedbuildjs/"):
        dest = f"/home/attah/Documents/speedbuildjs/{dest}"

    code = ""
    count = 1

    print("reading file ",dest)
    
    if os.path.exists(dest):
        with open(dest, "r") as file:
            data = file.read()
            lines = data.split("\n")
            for line in lines:
                code += f"{count} {line}\n"
                count += 1
            return code 
    else:
        print("does not exists")
        raise ValueError("File Does Not Exist")
    

def performUpdate(code:str,action:str,update:str,line:str):
    """
    Updates a given code string by performing add, remove, or replace actions on specified lines.

    Args:
        code (str): The original code as a string.
        action (str): The action to perform. Can be "add", "remove", or "replace".
        update (str): The code to add or replace. Ignored for "remove" action.
        line (str): The line number or range (e.g., "3" or "2-5") to apply the action.

    Returns:
        str: The updated code as a string.

    Notes:
        - For a line range (e.g., "2-5"), the action is applied to all lines in the range.
        - For a single line, the action is applied to that line.
        - "add" inserts the update code at the specified position.
        - "remove" deletes the specified line(s).
        - "replace" replaces the specified line(s) with the update code.
    """
    code_lines = code.split("\n")
    update_code = None if update == None else update.split("\n")

    if line != None:
        if "-" in line:
            new_code = []
            start, end = line.split("-")
            start = int(start)
            end = int(end)
            remove_line = []

            for index,line in enumerate(code_lines):
                lineNum = int(index) + 1
                if lineNum not in range(start,end+1):
                    new_code.append(line)
                else:
                    remove_line.append(line)
                    if action == "replace" and lineNum == end:
                        new_code.extend(update_code)
        else:
            pos = int(line)
            print("positomn ",pos, " ", len(code_lines))

            if action == "add":
                code_lines.insert(pos,"\n".join(update_code))
            else:
                if pos >= len(code_lines):
                    pos = -1

                code_lines.pop(pos)
                if action == "replace":
                    code_lines.extend(update_code)

            new_code = code_lines
            print("New code is == ",new_code)

        return "\n".join(new_code)

# Instead of write to file, lets call it update file

# Write File
def updateFile(filePath : str,fileName : str, action:str, update : str=None,line:str=None):

    """
    Updates a file at the specified path by performing the given action.

    Args:
        filePath (str): The directory path where the file is located.
        fileName (str): The name of the file to update.
        action (str): The action to perform on the file. Accepted values are "add", "replace", "remove", or "create".
        update (str, optional): The content to add or replace in the file. Required for "add", "replace", and "create" actions.
        line (str, optional): The specific line to target for the update. Used for "replace" or "remove" actions.
    
    Returns:
        str: An error message if the file is not found or the action is not recognized; otherwise, None.
    
    Notes:
        - For "add", "replace", and "remove" actions, the file must already exist.
        - For "create" action, a new file is created with the provided content.
        - The function relies on a helper function `performUpdate` to process the file content.
    """
    
    full_path = filePath
    if full_path.endswith("/") == False:
        full_path += "/"
    
    full_path += fileName
    
    if not full_path.startswith("/home/attah/Documents/speedbuildjs"):
        full_path = f"/home/attah/Documents/speedbuildjs/{full_path}" #remove this

    print("full path ",full_path)
    print(action,line)
    print("#"*10)
    print(update)
    print("#"*10)

    accepted_actions = ["add","replace","remove"]

    if os.path.exists(full_path):
        if action in accepted_actions:
            try:
                with open(full_path,"r+") as file:
                    code = file.read()
                    updated_code = performUpdate(code,action,update,line)
                    
                    file.seek(0)        # Move to the beginning of the file
                    file.truncate(0)    # Step 2: Clear the file
                    file.write(updated_code)

            except (FileNotFoundError, FileExistsError):
                return f"Could not find file {full_path}"
        else:
            return "Action not recognised; action could not be carried out"
    else:
        if action == "create":
            os.makedirs(full_path,exist_ok=True)
            with open(full_path,"w") as file:
                file.write(update)

    return "File Update was successful"

# Execute Command
def executeCommand(command : str, framework:str):
    """
    Executes a given shell command using the PythonExecutor.

    Args:
        command (str): The shell command to execute as a string.
        framework (str) : django or express

    Behavior:
        - Splits the command string into a list of arguments.
        - Determines if the command should be self-exiting based on whether the last argument is "runserver".
        - Executes the command in a specified working directory and environment.
        - Return the standard output and standard error of the executed command.
    """

    print("executing command ",command,"\n\n\n\n")

    home = str(Path.home())
    env = None
    wkdir = None

    if framework == "django":
        wkdir = f"{home}/.sb/environment/django/speedbuild_project"
        venv_path = f"{home}/.sb/environment/django/venv"

        if sys.platform == "win32":
            python_path =  os.path.join(venv_path, "Scripts", "python.exe")
        else:
            python_path = os.path.join(venv_path, "bin", "python")

        # Modify environment to use the venv
        env = os.environ.copy()
        env['VIRTUAL_ENV'] = venv_path

        if sys.platform == "win32":
            env['PATH'] = f"{os.path.join(venv_path, 'Scripts')};{env['PATH']}"
        else:
            env['PATH'] = f"{os.path.join(venv_path, 'bin')}:{env['PATH']}"
    elif framework == "express":
        wkdir = f"{home}/.sb/environment/express" #TODO : change this to be dynamic
    else:
        return

    commandExecutor = PythonExecutor()
    command = command.split(" ")

    selfExiting = False if  command[-1] == "runserver" or command[-1] == "dev" else True

    stdout, stderr = commandExecutor.runCommand(command=command,cwd=wkdir,env=env,self_exit=selfExiting)

    stdout = "\n".join(stdout)
    stderr = "\n".join(stderr)

    return [stdout,stderr]

    if len(stderr) > 0:
        return f"ERROR Message : {stderr}"
    
    if len(stdout) > 0:
        return f"Standard Output : {stdout}"


# if __name__ == "__main__":
#     file_path = "/home/attah/Documents/sb_final/speedbuild/"
#     file_name = "views.py"n

#     update = """def hello():
#     print("Say Buhari")
#     print("Tinubu go kill us")
#     print('RIP Buhari')"""
    
#     update = "print('RIP Buhari')"

#     # updateFile(file_path,file_name,"remove",None,"14-19")
#     # print(getFileContent(file_path,file_name))
#     executeCommand("python manage.py runserver")


system_prompt = """
You are a **smart AI agent** specialized in **Django projects**. You are part of a developer tool called **SpeedBuild**, which helps engineers extract reusable features from codebases and adapt them to new projects using natural language commands.

Each feature may span multiple files and depend on shared code. Your job is to **intelligently adapt a feature based on the developer’s instruction**, maintaining clean structure, preserving project architecture, and ensuring code quality.

---

### Your Responsibilities:

1. **Understand the developer's goal** by using the provided `task_description`.
2. **Process dependencies first** — helper functions, models, or utilities the feature depends on.
3. **Request any missing file** you need to continue the task, by name (with folder path if needed).
4. **Apply minimal, targeted changes** to the code while preserving existing structure.
5. **Include clear comments** only when they improve understanding of *why* something was changed.
6. **Return a structured JSON object** as output for all actions.
7. **Avoid generating redundant or boilerplate code unless explicitly required.**
8. **Never guess file contents** — always request them if unsure.
9. **Be concise, safe, and explicit**.

---

### Input Context (what you receive each time):

You are provided with:
- `task_description`: A plain English instruction from the developer (e.g., "Add email verification to user signup")
- `file_list`: A list of available project files you can request for processing (with paths)
- Optionally: the contents of one or more files.

---
]
### JSON Output Schema

You must always respond using this format:

```json
{
  "status": "success" | "request_file" | "need_clarification" | "error",
  "action": "Short description of what you're doing (e.g. 'Modified signup view')",
  "file_name": "Name of file you’re working on or requesting (with path if needed)",
  "content": "Modified code (if status is 'success') or null",
  "comments": "Brief, helpful explanation (why the change was made, or what is unclear)"
}
````

---

### Status Values

* `"request_file"` → You need a file's contents to continue
* `"success"` → You’ve successfully modified or created a file
* `"need_clarification"` → You need more info or the instruction is unclear
* `"error"` → You encountered a blocking issue (e.g. circular logic, incompatible structure)

---

When modifying a file, you must return a list of code block instructions inside the `content` field. 
Each item must contain:

```json
{
  "action": "add" | "replace" | "remove",
  "new_code": "<code to add or replace with>",
  "old_code": "<only for 'replace' or 'remove'>",
  "comment": "A short explanation of the change"
}
```

* If `action` is `"add"`: Only `new_code` is required.
* If `action` is `"replace"`: Both `new_code` and `old_code` are required.
* If `action` is `"remove"`: Only `old_code` is required.

---

### Examples

#### Requesting a File:

```json
{
  "status": "request_file",
  "action": "Requesting models.py to locate User model",
  "file_name": "accounts/models.py",
  "content": null,
  "comments": "Need this file to refactor the custom user logic."
}
```

#### Successful Code Change (Modification or File creation):


#### Modifying a File

```json
{
  "status": "success",
  "action": "Added 'Chef' model and updated 'Recipe' model to include 'prepared_by'",
  "file_name": "sb_app/models.py",
  "content": [
    {
      "action": "add",
      "new_code": "class Chef(models.Model):\n    name = models.CharField(max_length=100)\n    bio = models.TextField(blank=True)\n\n    def __str__(self):\n        return self.name",
      "comment": "Added a new 'Chef' model"
    },
    {
      "action": "replace",
      "old_code": "class Recipe(models.Model):\n    name = models.CharField(max_length=100)\n    directions = models.TextField()",
      "new_code": "class Recipe(models.Model):\n    name = models.CharField(max_length=100)\n    directions = models.TextField()\n    prepared_by = models.ManyToManyField(Chef, blank=True)",
      "comment": "Linked Recipe to Chef using ManyToManyField"
    }
  ],
  "comments": "Added a new Chef model and linked it to the Recipe model for tracking who prepared each recipe."
}
```

---

#### Deleting a File

```json
{
  "status": "delete",
  "action": "Remove unused test file",
  "file_name": "tests/temp_feature.py",
  "comments": "This file was used for temporary debugging and is no longer needed."
}
```

---

#### No Change Needed

```json
{
  "status": "copy",
  "action": "No changes required",
  "file_name": "utils/helpers.py",
  "comments": "The helper functions are already compatible with the new feature."
}
```

---

#### Ending Session

```json
{
  "status": "done",
  "action": "All tasks completed",
  "file_name": "",
  "content": "",
  "comments": "Reviewed and updated models.py, admin.py, and views.py. Registered new model and updated checkout logic."
}
```
---

#### Asking for Clarification:

```json
{
  "status": "need_clarification",
  "action": "Unclear how to integrate email verification",
  "file_name": null,
  "content": null,
  "comments": "Should the verification use Django's default email backend, or a third-party service like SendGrid?"
}
```

---

### Golden Rules

* Don’t hallucinate code. Request what you need.
* Prioritize correctness, safety, and architectural soundness.
* Only comment where it adds value.
* Be surgical, not verbose.
* Always follow the `task_description`.

---

"""



import os
import sys
import json
from pathlib import Path

user_home = str(Path.home())
config_path = f"{user_home}/.sb/config.json"

def get_sb_config():
    if not os.path.exists(config_path):
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        with open(config_path,"w") as file:
            file.write("{}")

        return {}

    with open(config_path,"r") as file:
        data = json.loads(file.read())
        return data

def save_config(config):
    if not os.path.exists(config_path):
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        
    with open(config_path, "w") as file:
        json.dump(config, file, indent=4)

def get_config():
    config = get_sb_config()
    # print(json.dumps(config, indent=4))
    return config

def setAPIKey(key_type,api_key):
    config = get_sb_config()

    keys_dictionary = {}
    if "llm_keys" in config.keys():
        keys_dictionary = config['llm_keys']

    if len(api_key) == 0 and key_type in keys_dictionary.keys():
        del keys_dictionary[key_type]
    else:
        keys_dictionary[key_type] = api_key

    
    config['llm_keys'] = keys_dictionary

    save_config(config)

def updateDefaultLLMModel(model_name):
    config = get_sb_config()
    config['default_model'] = model_name
    save_config(config)

def updateAuthToken(token_value):
    config = get_sb_config()

    keys_dictionary = {}
    if "jwt_tokens" in config.keys():
        keys_dictionary = config['jwt_tokens']

    if len(token_value) == 0:
        del config['jwt_tokens']
    else:
        config['jwt_tokens'] = token_value

    save_config(config)


import os
import re
import sys
import ast
import site
import importlib

import django
from django.conf import settings

from ....parsers.python.parser import PythonBlockParser
from .feature_dependencies import get_if_blocks, getCodeBlockNameAndType, remove_multiline_comments
from ....utils.utils import findFilePath


django_defaults = [
    "BASE_DIR",
    "SECRET_KEY",
    "DEBUG",
    "ALLOWED_HOSTS",
    "INSTALLED_APPS",
    "MIDDLEWARE",
    "ROOT_URLCONF",
    "TEMPLATES",
    "WSGI_APPLICATION",
    "DATABASES",
    "AUTH_PASSWORD_VALIDATORS",
    "LANGUAGE_CODE",
    "TIME_ZONE",
    "USE_I18N",
    "USE_TZ",
    "STATIC_URL",
    "DEFAULT_AUTO_FIELD"
]

default_django_apps = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]

class ManageDjangoSettings():
    def __init__(self,settings_path, venv_path):
        self.venv_path = venv_path
        self.settings_path = settings_path
        self.parser = PythonBlockParser()
        self.blocks = []
        self.blocks_without_comments = []
        self.code_mappings = {}

    def load_django_settings(self):
        """Loads Django settings dynamically from a given settings.py path."""
        settings_dir = os.path.dirname(self.settings_path)
        settings_module_name = "custom_settings"

        sys.path.insert(0, settings_dir)  # Add settings directory to Python path

        spec = importlib.util.spec_from_file_location(settings_module_name, self.settings_path)
        settings_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings_module)

        installed_packages = []

        if not settings.configured:  # Only configure if settings aren't already set
            settings.configure(
                INSTALLED_APPS=installed_packages,
                **{
                    key: getattr(settings_module, key)
                    for key in dir(settings_module) if key.isupper() and key != "INSTALLED_APPS"
                }
            )

        django.setup()  # Initialize Django

        return settings_module
    
    def getAllSourceProjectApps(self,source_apps):
        self.allSourceProjectApps = ast.literal_eval(source_apps.split("=", 1)[1].strip())

    def getTemplateMiddleWare(self,apps):

        try:
            MIDDLEWARE = self.code_mappings["MIDDLEWARE"]
        except KeyError:
            self.getSettingsVariablesName()
            MIDDLEWARE = self.code_mappings["MIDDLEWARE"]
            
        MIDDLEWARE = ast.literal_eval(MIDDLEWARE.split("=", 1)[1].strip())

        template_middleware = []

        for middleware in MIDDLEWARE:
            start = middleware.split(".")[0]
            if start in apps:
                template_middleware.append(middleware)

        return [template_middleware,MIDDLEWARE]


    def getSettingsVariablesName(self,get_code=False):
        with open(self.settings_path,"r") as file:
            data = file.read()
            data_without_multiline_comment = remove_multiline_comments(data)
            chunks = self.parser.parse_code(data)
            self.blocks = chunks

            setting_imports = []

            self.blocks_without_comments = self.parser.parse_code(data_without_multiline_comment)

            var_names = []
            var_name_to_code = {}
            all_variables = {}

            
            for chunk in chunks:
                if not (chunk.startswith("import ") or chunk.startswith("from ")) and not chunk.startswith("#") and not chunk.startswith('"'):
                    try:
                        var_name, var_type = getCodeBlockNameAndType(chunk)


                        if var_name == "MIDDLEWARE" or var_name not in django_defaults:
                            var_names.append(var_name)
                            var_name_to_code[var_name] = chunk

                        elif var_name == "INSTALLED_APPS":
                            self.getAllSourceProjectApps(chunk)

                            # print(var_name ,"is of type ",var_type)

                        if var_type == "if_statement":
                            # pass
                            condtional_statements = get_if_blocks(chunk)
                            for condition in condtional_statements:
                                body = condition['body']
                                if body:
                                    body = body.split("\n")
                                    for line in body:
                                        new_var_name, _ = getCodeBlockNameAndType(line)
                                        all_variables[new_var_name] = chunk
                                        # print("line ",new_var_name)

                                        if new_var_name not in django_defaults:
                                            var_names.append(new_var_name)
                                            var_name_to_code[new_var_name] = chunk

                                # print(body)
                        else:
                            all_variables[var_name] = chunk
                    except:
                        pass
                    # var_name, var_type = getCodeBlockNameAndType(chunk)

                    # print(var_type, " ",chunk[:20])

                    # if var_name == "MIDDLEWARE" or var_name not in django_defaults:
                    #     var_names.append(var_name)
                    #     var_name_to_code[var_name] = chunk

                    # all_variables[var_name] = chunk
                    # # print(var_name, " ")

                elif chunk.startswith("import ") or chunk.startswith("from "):
                    setting_imports.append(chunk)

            self.custom_conf = var_names
            self.code_mappings = var_name_to_code
            self.all_variables = all_variables

            self.setting_imports = setting_imports
            return var_names

    def use_venv(self):
        """Modifies sys.path to use the site-packages from the specified virtual environment."""
        python_version = f"python{sys.version_info.major}.{sys.version_info.minor}"
        
        if os.name == "nt":  # Windows
            site_packages = os.path.join(self.venv_path, "Lib", "site-packages")
        else:  # Linux/macOS
            site_packages = os.path.join(self.venv_path, "lib", python_version, "site-packages")

        if not os.path.exists(site_packages):
            raise FileNotFoundError(f"Site-packages directory not found in {self.venv_path}")

        sys.path.insert(0, site_packages)  # Prioritize this venv


    def get_app_paths(self,external_apps):
        """Returns the filesystem paths of third-party installed Django apps (excluding custom apps)."""

        # Ensure imports come from the specified virtual environment
        self.use_venv()

        app_paths = {} #get_venv_site_packages("/home/attah/jannis/venv")#
        site_packages = site.getsitepackages()  # Get site-packages directories

        def get_module_path(module_name):
            spec = importlib.util.find_spec(module_name)
            if spec and spec.origin:
                return os.path.dirname(spec.origin)
            return None
        
        for app in external_apps:
            try:
                module = importlib.import_module(app)
                app_path = os.path.dirname(module.__file__)

                app_paths[app] = app_path
            except ImportError:
                print(f"Warning: Could not import {app}")
            except TypeError:
                site_packages_path = os.path.join(self.venv_path, "lib", f"python{sys.version_info.major}.{sys.version_info.minor}", "site-packages")
                if os.path.exists(site_packages_path):
                    module_path = os.path.join(site_packages_path, app)
                    if os.path.exists(module_path):
                        app_paths[app] = module_path


        return app_paths

    def extract_words_from_code(self,code):
        # Extract words using regex (identifiers, keywords, function names, variable names)
        words = re.findall(r'[a-zA-Z_][a-zA-Z0-9_]*', code)
        
        return set(words)

    def find_settings_import(self,app_path,var_names):
        """Checks if 'from django.conf import settings' is in any file in the app."""
        use_settings = False
        configuration = set()

        for root, _, files in os.walk(app_path):
            for file in files:
                if file.endswith(".py"):
                    file_path = os.path.join(root, file)
                    with open(file_path, "r", encoding="utf-8") as f:
                        for line in f:
                            if "from django.conf import settings" in line:
                                # print(app_path, " ", file)
                                words = self.extract_words_from_code(f.read())
                                app_configuration = words.intersection(set(var_names))
                                configuration = configuration.union(app_configuration)
                                use_settings = True

        return [use_settings, list(configuration)]
    

    def getDjangoAppsConfigurations(self,apps):
        var_names = self.getSettingsVariablesName(self.settings_path)

        # print("here with apps ",var_names)

        # print("here with variable names ",var_names)
        conf_list = []

        self.load_django_settings()

        # apps.append("zoho_zeptomail")


        app_paths = self.get_app_paths(apps)

        # print("\nChecking for 'from django.conf import settings' in installed third-party apps...\n")
        for app, path in app_paths.items():
            use_setting, conf = self.find_settings_import(path,var_names) 
            if use_setting and len(conf) > 0:
                conf_list.extend(conf)
                # print(app," ",conf, " new conf")

        return conf_list
    

def get_project_settings(project_path):
    path = findFilePath(project_path,"asgi.py")
    if len(path) == 0:
        raise ValueError("not a django project, ensure you specified the right path")
    
    path = path[0]
    path = path.split("/")
    path.pop()
    path.append("settings.py")
    path = "/".join(path)

    settings_path = f"{project_path}/{path}"
    chunks = []

    with open(settings_path,"r") as file:
        data = file.read()
        chunks = PythonBlockParser().parse_code(data)

    # print(settings_path," path ")
    
    return [settings_path,chunks]

def updateMiddleware(temp_middleware,source,proj_middleware):
    if len(temp_middleware) > 0:
        for i in temp_middleware:
            index_in_source = source.index(i)
            insert_index = None
            if i not in proj_middleware:
                for x in range(index_in_source+1,len(source)):
                    if source[x] in proj_middleware:
                        insert_index = proj_middleware.index(source[x])
                        break
            
            if insert_index:
                proj_middleware.insert(insert_index, i)
            else:
                proj_middleware.append(i)
                
    return proj_middleware


import json
import importlib
import os

from .utils import findFilePath
from ..src.Django.utils.venv_utils import get_activated_venv
from ..src.Django.utils.django_app_dependencies import getDjangoAppsPackage

def is_standard_library(module_name):
    return importlib.util.find_spec(module_name) is not None

def get_package_name(import_name,data):

    import_name = import_name.split(" ")[0]

    if import_name in  data.keys():
        return[import_name,"str"]
        
    for key in data:
        if import_name in data[key]:
            return [key,"str"]
    
    # could not find package
    return [import_name,"None"]
        
    # exclude_path = ['pip','dateutil', '__pycache__']

    # with open("sb_app_mapping_cache.json","r") as cacheFile:
    #     data = json.loads(cacheFile.read())
    #     import_name = import_name.split(" ")[0]

    #     if import_name in  data.keys():
    #         return[import_name,"str"]
            
    #     for key in data:
    #         if import_name in data[key]:
    #             return [key,"str"]
        
    #     # could not find package
    #     return [import_name,"None"]
    
def getPackageNameMapping(project_path):

    venv = get_activated_venv()
    findFilePath(project_path, "urls.py")

    # use conditional statement to check if file is found
    project_name = findFilePath(project_path, "asgi.py")

    if len(project_name) == 0:
        raise(ValueError("Cannot find django project settings file"))
    
  
    project_name = project_name[0].split("/")[0]
    settings_path = f"{project_name}/settings.py"
    settings_path = os.path.join(project_path, settings_path)

    template_django_apps = getDjangoAppsPackage(settings_path,venv)
    return template_django_apps


from langchain.chat_models import init_chat_model
from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph import START,END,StateGraph, MessagesState
from langchain_core.messages import HumanMessage, AIMessage, SystemMessage

from .pydantic_model import LLMOutput
from .tools import getFileContent,executeCommand,updateFile

from .system_prompt import system_prompt
from .express_system_prompt import system_prompt as express_system_prompt


class ExtendedMessagesState(MessagesState):
    formatted_output: dict = {}
    structured_response: LLMOutput = None


class SpeedBuildLLMGraph():
    def __init__(self,model,model_provider):
        llm = init_chat_model(model=model,model_provider=model_provider)
        llm_with_tools = llm.bind_tools([getFileContent,executeCommand,updateFile])
        self.llm = llm_with_tools #.with_structured_output(LLMOutput)

        self.first_request = True

        # create graph
        self.setup()

    # def format_node(state: ExtendedMessagesState):
    #     response = state["structured_response"]
    #     formatted = {
    #         "action": response.action,
    #         "write_action": response.write_action,
    #         "line": response.line,
    #         "code": response.code,
    #         "command": response.command,
    #         "file_path": response.file_path,
    #     }

    #     return {"formatted_output": formatted}
    
    def get_messages(self,user_input: str, is_start=False,framework=None):
        if is_start:
            self.first_request = False
            return [
                SystemMessage(content=system_prompt if framework == None else express_system_prompt), #TODO : make this dynamic
                HumanMessage(content=user_input)
            ]
        return [HumanMessage(content=user_input)]
    
    def debugger(self,state:MessagesState):
        # response = self.llm.invoke(state["messages"])

        # return {
        #     # TODO
        #     "messages": [AIMessage(content=response.setup)],
        #     "structured_response": response
        # }
        return {"messages": [self.llm.invoke(state["messages"])]}
    
    
    def process_llm_command(self,state:ExtendedMessagesState):
        "process and test command"
        return {
            "messages": state["messages"].append(HumanMessage(content="response.setup"))
        }
    
        # or return end
    
    def setup(self):
        # state = ExtendedMessagesState()
        builder = StateGraph(MessagesState)

        # nodes
        builder.add_node("debugger",self.debugger)
        builder.add_node("tools", ToolNode([getFileContent,executeCommand,updateFile]))

        # edges
        builder.add_edge(START,"debugger")
        builder.add_conditional_edges(
            "debugger",
            tools_condition,
        )
        builder.add_edge("tools","debugger")
        builder.add_edge("debugger",END)
        # builder.add_edge("tools","format_node")
        # builder.add_edge("format_node","process_command")
        # builder.add_edge("process_command", "llm_call")

        # builder.add_conditional_edges()

        # build graph
        self.graph = builder.compile()
        return True


    def run(self,query,framework=None):
        messages = self.get_messages(query,self.first_request,framework)
        response = self.graph.invoke({'messages':messages})

        return response


import os
import yaml
import zipfile
from pathlib import Path

from ....utils.utils import getCurrentDjangoFiles
from .extract import getTemplateFileNames
from .feature_dependencies import removeDuplicates
from ....parsers.python.parser import PythonBlockParser

from ....utils.sb_agent import agent
from ....utils.query_split import query_splitter
from .write_dependency import sortFile, writeToFile
from ....utils.state import clearProjectState, copyFileToState, getOrCreateProjectSBId, updateState

user_home = str(Path.home())

def read_yaml(file_path):
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)
    return data


def getAppFileContent(appName,fileName,project_path):
    """
    Reads a Python file, separates import statements from the rest of the code, and returns them as two lists.

    Args:
        appName (str): The name of the application (used to construct the file path).
        fileName (str): The path to the file to be read.
        project_path (str): The root path of the project.

    Returns:
        list: A list containing two elements:
            - imports (list): A list of import statements found in the file.
            - code (list): A list of code blocks that are not import statements.

    Note:
        This function uses PythonBlockParser().parse_code to split the file content into code chunks.
    """
    imports = []
    code = []

    fileToUpdate = fileName.split("/")[-1]
    filePath = f"{project_path}/{appName}/{fileToUpdate}"

    with open(fileName, "r") as file:
        data = file.read()
        data = PythonBlockParser().parse_code(data)
        for chunk in data:
            if chunk.startswith("import ") or chunk.startswith("from "):
                # individualImports = getIndividualImports(chunk)
                imports.append(chunk)
            else:
                code.append(chunk)

    return [imports,code]


def processFile(fileName,appName,project_path,template_path,feature_app_name=None,django_root=None,processed_file_path={},project_id=None,file_written=[],conflicts=[]):
    imports = []
    code = []

    #get new feature code
    with open(fileName,"r") as file:
        data = file.read()
        data = PythonBlockParser().parse_code(data)

        for chunk in data:
            if chunk.startswith("import ") or chunk.startswith("from "):
                # individualImports = getIndividualImports(chunk)
                imports.append(chunk)
            else:
                code.append(chunk)
    
    #get existing file content
    fileImports, fileCode = getAppFileContent(appName,fileName,project_path)

    # print("filename is this ", fileName)

    # old implementation
    fileCode.extend(code)
    fileImports.extend(imports)
    fileCode = removeDuplicates(fileCode)
    fileImports = removeDuplicates(fileImports)

    #merge current and new code import
    # fileImports.extend(imports)
    # fileImports = removeDuplicates(fileImports)

    """
    🧩 You're Laying the Right Foundation
    Here’s what your sketch is doing, in essence:

    Get all "current" code chunk names (fileCode).

    For each chunk in the new version (code):

    If its name already exists in fileCode → there's a potential conflict.

    Else → it's a new chunk and can be appended.

    Perfectly valid.
    """

    # #merge current and new code while checking for conflicts
    # current_chunk_names = [get_assigned_variables(chunk) for chunk in fileCode]

    # for chunk in code:
    #     chunk_name = get_assigned_variables
    #     if chunk_name in current_chunk_names:
    #         pass
    #         # add chunk to current with conflict marker
    #     else:
    #         fileCode.append(chunk)


    fileToUpdate = fileName.split("/")[-1]
    filePath = f"{project_path}/{appName}"

    "get all files in filePath"

    # TODO make this dynamic, get all python file in app folder
    django_files = getCurrentDjangoFiles(filePath)#["models.py","urls.py","views.py","admin.py"]

    if "sb_app" not in fileName: # also check if filename is custom django names else write file in sb_utils folder

        new_file_path = fileName.replace(template_path,"")
        if new_file_path.startswith("/"):
            new_file_path = new_file_path[1:]

        # print(new_file_path.split("/"), " my filename is this\n\n")

        filePath = filePath.split("/")

        write_to_project_root = len(new_file_path.split("/")) > 1
        # filePath.pop()

        if write_to_project_root:
            # print("update ", fileToUpdate)

            if fileToUpdate not in django_files:
                filePath.append("sb_utils")
                #  update import here
                # writing to sb utils
                new_imports = []
                for line in fileImports:
                    path,dependency = line.split("import ")
                    path = path.replace("from","").replace(".sb_utils","").strip()
                    if path.startswith("."):
                        path = path[1:]


                    if f"{path}.py" in django_files:
                        new_imports.append(f"from {appName}.{path} import {dependency}")
                        continue

                    if "sb_utils" in filePath:
                        line = line.replace(".sb_utils","")

                    new_imports.append(line)
                    # elif line.startswith("import "):
                    #     new_imports.append(line)
                    # else:
                    #     pass

                fileImports = new_imports
        else:
            if django_root:
                # print(" we dey here oo help us #########################")
                # print(fileName)
                # print(f"{project_path}/{django_root}")
                filePath = [project_path,django_root]
            # settings files
            # TODO files to add to the main django folder

        filePath = "/".join(filePath)
        

        if not os.path.exists(filePath):
            os.makedirs(filePath)

    
    # update imports to remove unnecessary imports
    new_imports = []
    for line in fileImports:
        if ".sb_utils" in line:
            path,dependency = line.split("import")
            path = path.replace("from","").strip()
            path = path.replace(".sb_utils.","").strip()
            # print("sb_utils path is ", path)

            if f"{path}.py" in django_files:
                if f"{path}.py" != fileToUpdate:
                    new_imports.append(f"from .{path} import {dependency}")
                continue

            elif path in processed_file_path.keys():
                main_path = processed_file_path[path]
                new_imports.append(f"from {main_path}.{path} import {dependency}")
                continue

        new_imports.append(line)
    fileImports = new_imports

    importAsString = "\n".join(fileImports)
    codeAsString = "\n".join(fileCode)
    fileContent = importAsString + "\n\n" + codeAsString

    # print("############################# new path is ",filePath)

    # TODO : start here

    dest = f"{filePath}/{fileToUpdate}"

    file_state_key = dest.replace(project_path,"")

    # print("file path is ",filePath)
    # print("file name is ", fileName)
    # print("state key is ", file_state_key)
    # print("file to update ", fileToUpdate)

    # state before write
    if project_id:
        if fileToUpdate not in file_written:
            # copy file ]to prev state
            copyFileToState(project_id,dest,"prev")

            action = "update"
            if not os.path.exists(dest):
                action = "create"

            # update state.json to state file updated
            updateState(project_id,action,file_state_key)

            # add to file_written
            file_written.append(fileToUpdate)

    conflicts = writeToFile(filePath,fileContent,fileToUpdate,conflicts)
    if fileToUpdate == "models.py":
        sortFile(f"{filePath}/{fileToUpdate}")

    # state after write
    if project_id:
        # copy file to last_sb_update state
        copyFileToState(project_id,dest,"last_sb_update")

    return [file_written,conflicts]


def getTemplateFilteredFiles(template_unpack_path,feature_name):
    root_files = []
    filtered_files = []
    files = getTemplateFileNames(template_unpack_path)
 
    for i in files:
        if i not in ["settings.py", f"sb_{feature_name}.yaml"] and i.endswith("md") == False:
            if len(i.split("/")) == 1:
                # add to root_files also add to filtered_files
                root_files.append(i)
                filtered_files.append(i)

            elif len(i.split("/")) > 1:
                filtered_files.append(i)

    filtered_files = sorted(filtered_files, key=lambda x: not x.startswith('.sb_utils')) 
    return [filtered_files,root_files]

        
def getFeatureFromTemplate(template_path,project_root,template_name,django_root,app_name):
    """
    1) Ask for which app to implement feature. -- done
    2) Ask for Customization prompt and customize code. -- done
    3) Copy code and save in the right files (do proper refrencing).
    """

    if app_name is None:
        app_name = input("Which django app do you want to implement feature in : ")
    
    app_path = f"{project_root}/{app_name}"

    conflicts = []


    # feature_name = template_path.split("/")[-1].split(".")[0]
    feature_name = [word for word in template_path.split("/") if ".zip" not in word][-1]

    # print("feature name is ",feature_name)

    
    # path_to_template = f"{project_root}/.sb/{template_name}"

    path_to_template = f"{user_home}/.sb/sb_extracted" #{template_name}"
    
    if not os.path.exists(path_to_template):
        os.makedirs(path_to_template, exist_ok=True)

    processed_file_path = {}

    # print("app path is ", app_path)

    if os.path.exists(app_path) and os.path.isdir(app_path):
        print("\n\n------- Generating Feature -------\n\n")

        print(f"Getting template from {template_path}")

        # # unpack template to .sb folder
        # sb_dir = f"{project_root}/.sb"
        # os.makedirs(sb_dir, exist_ok=True)

        # # clear old content from folder
        # clear_folder(sb_dir)

        template_unpack_path = f"{path_to_template}/{feature_name}"
        os.makedirs(template_unpack_path, exist_ok=True)

        # unpack
        # TODO : clear previous zip content, before unzipping again
        print("Unpacking template")
        with zipfile.ZipFile(template_path, 'r') as zip_ref:
            zip_ref.extractall(template_unpack_path)

        # Get and read template configuration file
        # feature_name = template_path.split("_")[-1].replace(".zip","").strip()
        feature_name = feature_name.replace("speed_build_","")
        template_root = f"{path_to_template}/{template_name}"
        config_file_path = f"{template_root}/sb_{feature_name}.yaml"
        data = read_yaml(config_file_path)

        feature_app_name = data['feature_file_path'].split("/")[0]

        # proceed_with_customization = False
        filtered_files = []
        root_files = []

        # if prompt is None:
        #     while True:
        #         customize = input("would you like to customize feature (yes or no[default]) : ")
        #         customize = customize.strip()

        #         if customize.lower() in ["yes","no","y","n"]:
        #             if customize.lower() == "yes":
        #                 prompt = input("Enter Customization prompt (be as detailed as possible) : \n")
        #             break
        #         else:
        #             print("Enter a valid response")
        #             continue

        # if prompt is not None:
        #     # if prompt is None:
        #     #     prompt = input("Enter Customization prompt (be as detailed as possible) : \n")
        #     # make agent call here
        #     # pass prompt and a list of all files in template
    
        #     # TODO : leave commented for now
        #     actions = query_splitter(prompt)

        #     filtered_files,root_files = getTemplateFilteredFiles(template_unpack_path,feature_name)

        #     for root_file in root_files:
        #         path_name = root_file.replace(".py","").strip()
        #         processed_file_path[path_name] = django_root 

        #     root_files.extend(filtered_files) # merge files NB, made sure root files appear first
        #     filtered_files = root_files

        #     # print("all actions are \n", actions)
        #     for step in actions:
        #         agent(filtered_files,step,app_path,feature_name)
        # else:
        #     pass

        files = getTemplateFileNames(template_unpack_path)

        for i in files:
            if i not in ["settings.py", f"sb_{feature_name}.yaml"] and i.endswith("md") == False:
                if len(i.split("/")) == 1:
                    root_files.append(i)

                elif len(i.split("/")) > 1:
                    filtered_files.append(i)

        filtered_files = sorted(filtered_files, key=lambda x: not x.startswith('.sb_utils')) 

        for root_file in root_files:
            path_name = root_file.replace(".py","").strip()
            processed_file_path[path_name] = django_root 

        root_files.extend(filtered_files) # merge files NB, made sure root files appear first
        filtered_files = root_files

        # TODO: print feature description, along with dependencies and fields

        # print(filtered_files)

        project_id = getOrCreateProjectSBId(project_root)

        # clear project state 
        clearProjectState(project_id)

        file_written = []

        # TODO : get filtered files again so u get new files llm created
        filtered_files,root_files = getTemplateFilteredFiles(template_unpack_path,feature_name)

        for file in filtered_files:
            file_path = f"{template_root}/{file}"
            # print(file_path)
            file_written,conflicts = processFile(file_path,app_name,project_root,template_root,feature_app_name,django_root,processed_file_path,project_id,file_written,conflicts)

    else:
        raise ValueError(f"No app with name {app_name} in {project_root}")
        return None
    
    return [app_name,template_root,conflicts]



import os
import ast
import zipfile
import tempfile

def extract_code_chunks(file_path):
    """Parse Python code and return sorted normalized code chunks (functions, classes, top-level)."""
    with open(file_path, 'r', encoding='utf-8') as f:
        source = f.read()

    try:
        tree = ast.parse(source)
    except SyntaxError:
        return []  # Return empty for non-Python files or syntax errors

    chunks = []

    for node in tree.body:
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            chunks.append(ast.unparse(node).strip())
        else:
            # Capture top-level code (e.g., assignments, expressions)
            chunks.append(ast.unparse(node).strip())

    # Normalize and sort to make order irrelevant
    normalized = sorted([chunk for chunk in chunks if chunk])
    return normalized

def compare_code_files(file1, file2):
    """Compare two Python code files ignoring chunk order but ensuring chunk content matches."""
    chunks1 = extract_code_chunks(file1)
    chunks2 = extract_code_chunks(file2)
    return chunks1 == chunks2


def compare_js_code_files(file1, file2):
    """TODO : Compare two javascript/typescript code files ignoring chunk order but ensuring chunk content matches."""
    # chunks1 = extract_code_chunks(file1)
    # chunks2 = extract_code_chunks(file2)
    # return chunks1 == chunks2


def compare_zip_code_folders(zip1_path, zip2_path):
    """Compare two zipped Python code folders with chunk-level matching (order-independent), skipping sb_*.yaml files."""
    with tempfile.TemporaryDirectory() as dir1, tempfile.TemporaryDirectory() as dir2:
        # Extract both zips
        with zipfile.ZipFile(zip1_path, 'r') as zip1:
            zip1.extractall(dir1)
        with zipfile.ZipFile(zip2_path, 'r') as zip2:
            zip2.extractall(dir2)

        # Build relative file list from dir1
        for root, _, files in os.walk(dir1):
            for filename in files:
                rel_path = os.path.relpath(os.path.join(root, filename), dir1)
                file1 = os.path.join(dir1, rel_path)
                file2 = os.path.join(dir2, rel_path)

                # Skip files that start with sb_ and end with .yaml
                if filename.startswith('sb_') and filename.endswith('.yaml'):
                    continue

                # Check existence
                if not os.path.exists(file2):
                    print(f"Missing file in second zip: {rel_path}")
                    return False

                # For .py files, do chunk comparison
                if filename.endswith("py"):
                    if not compare_code_files(file1, file2):
                        print(f"Python file mismatch: {rel_path}")
                        return False
                # elif filename.split(".")[-1] in ["js","ts"]:
                #     if not compare_js_code_files(file1, file2):
                #         return False
                else:
                    # For other files, compare exact content
                    with open(file1, 'rb') as f1, open(file2, 'rb') as f2:
                        if f1.read() != f2.read():
                            print(f"Binary or non-Python file mismatch: {rel_path}",file1,file2)
                            return False

        # Check for extra files in second dir
        for root, _, files in os.walk(dir2):
            for filename in files:
                rel_path = os.path.relpath(os.path.join(root, filename), dir2)
                file1 = os.path.join(dir1, rel_path)

                # Skip files that start with sb_ and end with .yaml
                if filename.startswith('sb_') and filename.endswith('.yaml'):
                    continue

                if not os.path.exists(file1):
                    print(f"Extra file in second zip: {rel_path}")
                    return False

        return True


import json
from typing import Optional
from pydantic import BaseModel, Field


class DebuggerOutput(BaseModel):
    """
    {
        action:"read file | update file | run command | end",
        description:"",
        file_path : "" | none,
        command : "" | None,
        file_action : "add | replace | remove"
        file_lines : "1(single line ) | 10-20 (multi line)" | None
        new_code:""
    }
    """
    action: str = Field(description="Action to perform")
    description: str = Field(description="Action to perform")
    file_path: Optional[str] = Field(description="Action to perform", default=None)
    command: Optional[str] = Field(description="Action to perform", default=None)
    file_action: Optional[str] = Field(description="Action to perform", default=None)
    file_lines: Optional[str] = Field(description="Action to perform", default=None)
    new_code: Optional[str] = Field(description="Action to perform", default=None)


def formatLLMOutput(data : DebuggerOutput, stringify=False):
    formatted =  {
        "action":data.action,
        "description":data.description,
        "file_path":data.file_path,
        "command":data.command,
        "file_action":data.file_action,
        "file_lines":data.file_lines,
        "new_code":data.new_code
    }

    if stringify:
        return json.dumps(formatted)
    
    return formatted


import json
import time
import random
import string
import websockets
import webbrowser

from .cli_utils import updateAuthToken
from ..utils.server_address import ws_address, app_address

def caculateTimeDifference(past_time):
    current_time = time.time()
    diff_in_seconds = current_time - past_time
    diff_in_minutes = diff_in_seconds // 60

    return diff_in_minutes

def getLocalUserData():
    try:
        with open(".probe/userData.json","r") as userDataFile:
            data = json.loads(userDataFile.read())
            return data
    except FileNotFoundError:
        raise UserWarning("You need to login to proceed\nrun 'probe login'")

def saveUserDataToLocalStorage(data):
    with open(".probe/userData.json","w") as userDataFile:
        json.dump(data, userDataFile, indent=4)

def generate_ws_token(length=6):
    # Define the possible characters (letters and digits)
    characters = string.ascii_letters + string.digits
    # Randomly choose `length` characters from the set
    random_string = ''.join(random.choice(characters) for _ in range(length))
    # Add current timestamp (in seconds) to make it more unique
    timestamp = str(int(time.time()))
    return f"{random_string}_{timestamp}"

async def openAuthPage(code,page="login"):
    if page == "login":
        address = f"{app_address}login?token={code}"
    else:
        address = f"{app_address}register?token={code}"

    webbrowser.open(address)
    response = await connect_to_websocket(f"{ws_address}chat/{code}/")
    # print(response)
    userData = {
        "access_token":response['message']['access'],
        "refresh_token":response['message']['refresh'],
        "time_created":time.time()
    }

    updateAuthToken(userData)

    # saveUserDataToLocalStorage(userData)
    print("Login Successful")


async def connect_to_websocket(WEBSOCKET_URI,data=None):
    headers = {
       "Origin": "https://backend.speedbuild.dev",  # Same origin as the server
    }

    async with websockets.connect(WEBSOCKET_URI,additional_headers=headers) as websocket:
        # print("Connected to WebSocket server")

        # data = {"message":"Hello, WebSocket!"}

        if data != None:
            # Send a message
            await websocket.send(json.dumps(data))
            print(f"Message sent: {data}, WebSocket!")

        # Keep listening for messages until "terminate" is received
        while True:
            response = await websocket.recv()
            print(f"Message received: {response}")
            data = json.loads(response)
            
            # Check if the received message contains the termination command
            if data['message']['status'].lower() == "terminate":
                # print("Termination message received. Closing connection.")
                return data
            
            # elif data['message'].lower() == "notify":
            #     print("notifying")


# remove blank lines before splitting
# blank line is breaking the code
import ast
import re

def has_unclosed_multiline_comment(code: str) -> bool:
    triple_quote_pattern = re.compile(r"('''|\"\"\")")

    # Remove escaped triple quotes and comments
    code = re.sub(r'\\("""|\'\'\')', '', code)
    code = re.sub(r'#.*', '', code)  # Remove inline comments

    # Find all triple quote occurrences
    quotes = triple_quote_pattern.findall(code)
    # print(len(quotes))

    # Count of triple quotes must be even for all to be closed
    return len(quotes) % 2 != 0

class PythonBlockParser:
    def seperateImport(self,code):
        import_lines = []
        tree = ast.parse(code)
        for node in ast.iter_child_nodes(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    import_lines.append(f"import {alias.name}")
            elif isinstance(node, ast.ImportFrom):
                module = node.module or ""
                for alias in node.names:
                    import_lines.append(f"from {module} import {alias.name}")

        return import_lines
    
    def chunkHasMultiLineComment(self,chunk):
        if '"""' in chunk or "'''" in chunk:
            return True
        return False
    
    def mergeMultiLineComment(self, chunks):
        add_to_previous = False
        memory = []
        sorted_chunks = []

        for chunk in chunks:
            memory.append(chunk)

            if self.chunkHasMultiLineComment(chunk):
                if has_unclosed_multiline_comment(chunk):
                    add_to_previous = not add_to_previous

            if not add_to_previous:
                # Only append when multiline comment block is fully closed
                sorted_chunks.append("\n".join(memory))
                memory = []  # Clear for next block

        # If memory still has content at the end (e.g. unclosed comment at end of file)
        if memory:
            sorted_chunks.append("\n".join(memory))

        return sorted_chunks


    
    def clean_parsed_code(self,chunks):
        chunks = self.mergeMultiLineComment(chunks)
        """
        Process and clean parsed code chunks by handling import statements and combining related code segments.

        This method processes a list of code chunks by:
        1. Separating multiple imports into individual import statements
        2. Combining consecutive chunks that belong together (except for decorator lines)
        3. Managing a temporary memory buffer for related code segments

        Args:
            chunks (list): A list of string chunks containing Python code segments

        Returns:
            list: A processed list of code chunks where:
                - Multiple imports are split into separate chunks
                - Related code segments are combined
                - Decorators are handled separately

        Example:
            For input chunks:
            ['import os, sys', '@decorator\n', 'def func():\n    pass']
            
            Returns:
            ['import os', 'import sys', '@decorator\ndef func():\n    pass']
        """
        processed_chunks = []
        memory = []
        add_to_previous = False

        for chunk in chunks:
            if chunk.startswith("import ") or chunk.startswith("from "):
                if chunk.count("import ") > 1:
                    for i in self.seperateImport(chunk):
                        if i.startswith("import "):
                            processed_chunks.extend(self.split_packed_imports(i))
                        else:
                            processed_chunks.append(i)
                else:
                    if chunk.startswith("import "):
                        processed_chunks.extend(self.split_packed_imports(chunk))
                    else:
                        processed_chunks.append(chunk)
                continue #start over

            memory.append(chunk)

            if not chunk.startswith("@"):
                processed_chunks.append("\n".join(memory))
                memory = [] #clear memory

        return processed_chunks
    
    def split_packed_imports(self,chunk):
        """
        Splits packed import statements into individual import statements.
        This method processes a chunk of Python code containing multiple imports 
        separated by commas and converts them into separate import statements.
        Args:
            chunk (str): A string containing one or more import statements potentially
                         separated by commas.
        Returns:
            Union[str, List[str]]: If the chunk starts with '#', returns the original chunk.
                                  Otherwise, returns a list of individual import statements.
        Examples:
            >>> split_packed_imports("import os, sys")
            ['import os', 'import sys']
            >>> split_packed_imports("# import comment")
            '# import comment'
        """
        if chunk.startswith("#"):
            return chunk
        
        imports = ["import "+pack.strip() for pack in chunk.replace("import ","").split(",")]
        return imports
    
    def parse_code(self, code_string):
        """Parse Python code into top-level blocks."""
        lines = code_string.split('\n')
        
        blocks = []
        current_block = []
        stack = []  # For tracking delimiters (), [], {}
        
        i = 0
        while i < len(lines):
            line = lines[i].rstrip()
            
            # Skip empty lines between blocks
            if not line.strip():
                if current_block:
                    blocks.append('\n'.join(current_block))

                    current_block = []
                i += 1
                continue
            
            # Calculate indentation
            indent = len(line) - len(line.lstrip())
            
            # Start of a new block
            if not current_block:
                current_block = [line]
                stack = self._update_delimiter_stack(line, [])
                i += 1
                continue
            
            # If we're at the top level (no indent) and there's no open delimiters
            if indent == 0 and not stack and not self._is_incomplete_line(current_block[-1]):
                stripped = line.lstrip()
                # If this is a new top-level construct, start a new block
                if not (stripped.startswith(('elif', 'else', 'except', 'finally'))):
                    blocks.append('\n'.join(current_block))
                    current_block = [line]
                    stack = self._update_delimiter_stack(line, [])
                else:
                    # Continue the current block for elif/else/except/finally
                    current_block.append(line)
            else:
                # Add to current block
                current_block.append(line)
                stack = self._update_delimiter_stack(line, stack)
            
            i += 1
        
        # Add the last block if exists
        if current_block:
            blocks.append('\n'.join(current_block))
        
        if len(blocks) == 0:
            return []
        
        new_blocks = [blocks[0]]

        for i in range(1,len(blocks)):
            chunk = blocks[i]
            # print(f"line {chunk}\n\n")
            stack = []
            lines = chunk.split("\n")
            # print(type(chunk)," ",chunk)

            for line in lines:
                stack,has_close_but_no_opening = self._has_close_but_no_opening(line.strip(),stack)


            if chunk.startswith("    ") or has_close_but_no_opening:
                new_blocks[-1] += f"\n\n{chunk}"
            else:
                new_blocks.append(chunk)

        return self.clean_parsed_code(new_blocks)
    
    def _update_delimiter_stack(self, line, stack):
        """Update the stack of delimiters for the line."""
        delimiters = {'(': ')', '[': ']', '{': '}'}
        
        for char in line:
            if char in delimiters:
                stack.append(char)
            elif char in delimiters.values():
                if stack and delimiters[stack[-1]] == char:
                    # print("popping ",line)
                    stack.pop()
        
        return stack
    
    def _has_close_but_no_opening(self, line, stack):
        """Update the stack of delimiters for the line."""
        delimiters = {'(': ')', '[': ']', '{': '}'}
        has_close_but_no_opening = False

        for char in line:
            if char in delimiters:
                stack.append(char)
            elif char in delimiters.values():
                if stack and delimiters[stack[-1]] == char:
                    # print("popping ",line)
                    has_close_but_no_opening = False
                    stack.pop()
                else:
                    has_close_but_no_opening = True
        
        return [stack,has_close_but_no_opening]
    
    def _is_incomplete_line(self, line):
        """Check if the line appears incomplete (e.g., ends with operator)."""
        operators = ['=', '+', '-', '*', '/', '%', '&', '|', '^', ',0', '\\']
        stripped = line.rstrip()
        return any(stripped.endswith(op) for op in operators)


def unindentCode(code):
    block = []
    lines = code.split("\n")
    start = lines[0]
    indent = 0

    for char in start:
        if char == " ":
            indent += 1

    for line in lines:
        while len(line[:indent].strip()) > 0:
            indent -= 1
        # print(line[:indent], "data")
        block.append(line[indent:].strip())
    
    return "\n".join(block)

def indentCode(code, indent=4):
    block = []
    lines = code.split("\n")
    # start = lines[0]

    for line in lines:
        block.append(f"{" "*indent}{line}")
    
    return "\n".join(block) 
    


from typing import List
from pydantic import BaseModel, Field
from langchain.chat_models import init_chat_model
from langchain_core.messages import HumanMessage, SystemMessage

system_prompt = """
You are an intelligent parser agent for an application customization service.

Your task:
- Analyze the user's request.
- Break it down into a list of atomic, actionable sub-prompts.

Rules:
- Atomic means: each sub-prompt should describe a single self-contained code mutation (e.g., add, remove, edit, modify).
- Do not decompose into micro-steps like opening files, identifying code, or locating fields.
- Do not solve, execute, or explain any sub-prompt.
- Preserve original terminology from the user’s input when possible.
- Do not include introductions, explanations, or summaries. Only output the list.
- Actions should be phrased as high-level code mutations, not developer actions.
- Ignore any action involving database schema; focus only on code logic changes.
"""

class LLMOutput(BaseModel):
    prompts : List[str] = Field(description="List of sub prompts")

def splitPrompt(prompt):
    model = init_chat_model(model="gpt-4o", model_provider="openai")
    model = model.with_structured_output(LLMOutput)

    res = model.invoke([
        SystemMessage(content=system_prompt),
        HumanMessage(content=prompt)
    ])

    return res.prompts


import os
import zipfile
from pathlib import Path

from ....cli_output.output import StatusManager
from ....langgraph_agent.agents.workflow.tools import executeCommand
from ....langgraph_agent.agents.workflow.claude_workflow import SpeedBuildWorkflow


home = str(Path.home())
test_environment_path = f"{home}/.sb/environment/express"

def getTemplateFileNames(path):
    file_paths = []
    for root, _, files in os.walk(path):
        if "node_modules" in root:
            continue
        for file in files:
            # Get the absolute path by joining root and file
            absolute_path = os.path.join(root, file)
            file_paths.append(absolute_path)
    return file_paths


def debug_and_customize_template_code(command,prompt=None):

    """
    Debug and customize template code by analyzing command output and errors.
    This function creates a debugging workflow to analyze and fix issues with a given command.
    It uses the SpeedBuildWorkflow debugger to iteratively analyze errors and provide debugging steps.

    Args:
        command (str): The shell command to debug and customize
        prompt (str, optional): Additional prompt/context for code customization. Defaults to None.

    Returns:
        bool: True if debugging completed successfully, False if debugging ended without resolution

    Note:
        - Uses SpeedBuildWorkflow debugger with GPT-5 model
        - Captures command output and error messages
        - Creates debugging context with original command, output and errors
        - Runs an iterative debugging workflow until resolution or failure
    """

    llm_agent = SpeedBuildWorkflow("gpt-4o","openai","express")

    if prompt == None: # normal debugging
        # Get initial error and command output
        _,stdout,stderr = llm_agent.run_shell_command(command)

        if len(stderr.strip()) == 0:
            # return if there is no error to debug
            return
        
        llm_agent.last_error = stderr 
        llm_agent.last_command_output = stdout
        llm_agent.original_command = command

        # Create workflow for debugging a specific command
        llm_agent.create_debugger_workflow()

        llm_agent.run()

        if not llm_agent.success:
            print("workflow ended without resolution")
            return False
            
    else: # customization workflow here
        llm_agent.create_debugger_workflow("customization")

        print("processing prompt : ",prompt)
        project_files = getTemplateFileNames(test_environment_path)
        llm_agent.last_command_output = f"""
        customization prompt : **{prompt}**
        project files : {project_files}
        """

        llm_agent.run()

        if not llm_agent.success:
            print("workflow ended without resolution")
            return False
        
    print("completed successfully!")
    return True


def debugTemplate(path_to_template):
    logger = StatusManager()
    home = str(Path.home())

    test_environment_path = f"{home}/.sb/environment/express"

    if os.path.exists(test_environment_path):
        os.makedirs(test_environment_path,exist_ok=True)

    # extract and move template code into test environment
    logger.start_status("Unpacking template")
    with zipfile.ZipFile(path_to_template, 'r') as zip_ref:
        zip_ref.extractall(test_environment_path)
        logger.stop_status("Unpacking Completed")

    # # Install node dependencies
    logger.start_status("Installing package dependencies in test environment")
    _, error = executeCommand("npm install","express")

    if len(error) > 0:
        error = "\n".join(error)
        logger.stop_status()
        raise ValueError(f"Could not Install node dependencies, command executed with error : {error}")
    
    logger.stop_status("Dependencies Installed")

    # TODO : check response to see if debug_and_customize_template_code was successful

    # Test template code
    logger.print_message("Checking and debugging template code for errors")
    debug_and_customize_template_code("npm run dev")


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="speedbuild",
    version="0.1.6.2",
    description="Extracts, adapts, and deploys battle-tested features from existing codebases to new projects—complete with all dependencies, configurations, and framework integrations.",
    long_description=long_description,
    long_description_content_type="text/markdown",  # Required for README.md
    packages=find_packages(),
    install_requires=[
        "django",
        "pyyaml",
        "openai",
        "requests",
        "websockets",
        "langgraph",
        "langchain",
        "esprima",
        "rich",
        "langchain-openai"
    ],
    entry_points={
        'console_scripts': [
            'speedbuild=speedbuild.sb:start',
        ],
    },
)



import os

from ....parsers.javascript_typescript.jsParser import JsTxParser
from .extractJs import getFileNameFromDir
from .reverseImport import convertToImportStatements

def list_all_files(folder_path):
    file_paths = []
    for root, dirs, files in os.walk(folder_path):
        for filename in files:
            full_path = os.path.join(root, filename)
            file_paths.append(full_path)
    return file_paths

def getNewImportPath(file_path, dep_path):
    file_path = os.path.abspath(os.path.normpath(file_path))
    dep_path = os.path.abspath(os.path.normpath(dep_path))

    return os.path.relpath(dep_path,file_path)

async def updateFileImportInFile(filename,dependency,root_path):
    new_dict = {}
    parser = JsTxParser()
    file_path = "/".join(filename.split("/")[:-1])

    dependency = dependency.split("/")
    dependency_file = dependency[-1].lstrip()
    dependency_path = "/".join(dependency[:-1])

    _,chunks,_,import_deps = await parser.parse_code(filename,False,True)

    for key in import_deps.keys():
        full_file_path = getFileNameFromDir(dependency_path,dependency_file)
        
        if full_file_path is not None:
            dep = os.path.basename(full_file_path)
            key_base_name = os.path.basename(key)

            if key_base_name.endswith(dep) or key_base_name.endswith(dep.split(".")[0]):
                value = import_deps[key]
                new_key = getNewImportPath(file_path,dependency_path)
                new_key = f"{new_key}/{dependency_file}"
                new_dict[new_key] = value
            else:
                new_dict[key] = import_deps[key]
    
    # Update file here
    if new_dict.keys() != import_deps.keys():
        code = "\n".join(convertToImportStatements(new_dict))
        code += f"\n\n{"\n".join(chunks)}"

        with open(filename,"w") as file:
            file.write(code)

async def moveFile(file_path, folder):

    # check if file exists
    if not os.path.exists(file_path):
        return None
    
    # Move the file to the new location
    os.makedirs(folder, exist_ok=True)
    new_file_path = os.path.join(folder, os.path.basename(file_path))
    os.rename(file_path, new_file_path)

    # update file refrence in other files
    files = list_all_files("./output")
    for file in files:
        if file != file_path:
            await updateFileImportInFile(file,new_file_path,"./output")


# moveFile("./output/constants.js","./output/config")


import os
import sys
import asyncio
from pathlib import Path

from .auth.cli_utils import get_config

from .cli_output.output import StatusManager

from .auth.sb_user import sbAuth
from .auth.auth import manageAuth
from .auth.serverCall import pingServer

from .utils.utils import pullPythonPackageJSON
from .utils.pull_template import pullInitialTemplates
from .utils.package_utils import getPackageNameMapping

from .old_sb import createTemplate, implementFeature

from .src.undo import undoSBDeploy
from .src.Express.deploy.deployJS import deployExpressFeature
from .src.Express.extract.jsPath import startExpressExtraction

from rich.prompt import Prompt

def extractFeature(home_path,project_root,args):
    try:
        target = args[2]
        extract_from = args[3]
        framework = args[4]

        framework = framework.replace("--","").lower()

        # logger = StatusManager()

        if framework == "django":
            # Extract django feature
            # logger.start_status("Extracting django feature")
            project_name = os.path.basename(project_root)
            packageToNameMapping = getPackageNameMapping(project_root)
            createTemplate(extract_from, target,project_name,packageToNameMapping,project_root)
            
        elif framework == "express":
            #Extract Express feature
            asyncio.run(startExpressExtraction(target,extract_from,project_root))

    except IndexError:
        print("Usage : python final.py <what_to_extract> <extraction_entry_point> --framework")


def deployFeature(home_path,project_root,args):
    try:
        template = args[2]
        framework = args[3]
        framework = framework.replace("--","").lower()
        print(f"Deploying {template}")
        
        customization_request = Prompt.ask("[cyan bold] Will you like to customize this feature (yes/no) [/cyan bold]")
        while customization_request.lower().strip() not in ['yes','no']:
            customization_request = Prompt.ask("[cyan] Please enter a valid response (yes/no)[/cyan]")

        customization_request = customization_request.strip() == "yes"

        if framework == "django":
            implementFeature(template)
        elif framework == "express":
            asyncio.run(deployExpressFeature(template,project_root,customization_request)) #true is so we customize otherwise use false

    except IndexError:
        print("Usage : python final.py deploy <template_name>")

def listExtractedFeature(args):
    print("Listing All Extracted Templates")
    user_home = str(Path.home())
    templates_dir = f"{user_home}/.sb_zip"

    if not os.path.exists(templates_dir):
        os.makedirs(templates_dir)

    count = 1
    templates = os.listdir(templates_dir)

    for template in templates:
        if template.startswith("speed_build"):
            print(f"    {count}) {template}")
            count += 1
    print("\n")

def start():
    args = sys.argv
    home = str(Path.home())
    current_path = os.path.abspath(".")


    try:
        # TODO : remove this
        api_key = os.environ.get("OPENAI_API_KEY",None)
        if api_key == None:
            config = get_config()
            if "openai" in config['llm_keys'].keys():
                os.environ["OPENAI_API_KEY"] = config["llm_keys"]['openai']
    except:
        pass

    try:
        command = args[1]

        if command == "extract":
            # sb.py extract / views.py --express
            extractFeature(home,current_path,args)

        elif command == "deploy":
            # sb.py deploy template_name
            deployFeature(home,current_path,args)

        elif command == "list":
            listExtractedFeature(args)

        elif command == "undo":
            print("performing undo")
            undoSBDeploy()

        elif command == "auth":
            asyncio.run(sbAuth())

        elif command == "ping":
            pingServer()
            
        elif command == "setup":
            print("Seting up speedbuild")
            pullInitialTemplates()
            pullPythonPackageJSON()
            asyncio.run(manageAuth("register"))

    except IndexError as e:
        print("Available commands :\n- extract\n- deploy\n- list\n- undo\n- auth\n- ping\n- setup")